from functools import wraps, partial

from bacpypes.object import (
    AnalogInputObject,
    AnalogValueObject,
    BinaryValueObject,
    Property,
    register_object_type,
    registered_object_types,
    DatePatternValueObject,
)
from bacpypes.primitivedata import CharacterString, Date, Time, Real, Boolean, Integer
from bacpypes.basetypes import EngineeringUnits, BinaryPV, Polarity
from bacpypes.local.object import AnalogValueCmdObject, Commandable, MinOnOff

_SHOULD_BE_COMMANDABLE = ["relinquishDefault", "outOfService", "lowLimit", "highLimit"]

"""
Template

Decorators is an effort to handle object creation without explicitly declare a new class
depending on the properties or features required.



# Usage

## bacnet_properties
This decorator takes a dict as argument defining supplmental properties

## bacnet_property
This decorator takes a simple property and its default value, adds it to object

## Commandable
This decoratore will modify the base class and create a new class that inherit from _commando (see local.object.py)

## Add feature
This decorator works the same than commandable. Could serve as a way to add behaviour like MinOnOff, events, limits, etc...

## Example::

    properties = {"outOfService" : False,
                "relinquishDefault" : 0,
                "units": "degreesCelsius",
                "highLimit": 98}

    @bacnet_properties(properties)
    @commandable()
    def av(instance, objectName, presentValue, description):
        OBJECT_TYPE = AnalogValueObject
        return create(OBJECT_TYPE,instance, objectName, presentValue, description)

    @add_feature(MinOnOff)
    @commandable()
    def bv(instance, objectName, presentValue, description):
        OBJECT_TYPE = BinaryValueObject
        return create(OBJECT_TYPE,instance, objectName, presentValue, description)

    @commandable()
    def datepattern(instance, objectName, presentValue, description):
        OBJECT_TYPE = DatePatternValueObject
        return create(OBJECT_TYPE,instance, objectName, presentValue, description)

    ### The creation takes place when the functions are called
    a = av(1,'AnalogValueName',10,'AnalogValue Description')
    b = bv(1,'BV Name','inactive','BinaryValue Description')
    c = datepattern(1,'My Date Pattern',None,'DatePattern Description')

"""


def _allowed_prop(obj):
    allowed_prop = {}
    for each in type(obj).properties:
        allowed_prop[each.identifier] = each.datatype
    for base in type(obj).__bases__:
        try:
            for each in base.properties:
                allowed_prop[each.identifier] = each.datatype
        except AttributeError:
            pass
    return allowed_prop


def _mutable(property_name, force_mutable=False):
    if property_name in _SHOULD_BE_COMMANDABLE and not force_mutable:
        mutable = True
    elif force_mutable:
        mutable = force_mutable
    else:
        mutable = False
    return mutable


def make_commandable():
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            allowed_prop = _allowed_prop(obj)
            _type = allowed_prop["presentValue"]
            _commando = Commandable(_type)
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__ + "Cmd"
            new_type = type(base_cls_name, (_commando, base_cls), {})
            new_type.__name__ = base_cls_name
            register_object_type(new_type, vendor_id=842)
            objectType, instance, objectName, presentValue, description = args
            new_object = new_type(
                objectIdentifier=(base_cls.objectType, instance),
                objectName="{}".format(objectName),
                presentValue=presentValue,
                description=CharacterString("{}".format(description)),
            )
            return new_object

        return wrapper

    return decorate


def add_feature(cls):
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            base_cls = obj.__class__
            base_cls_name = obj.__class__.__name__ + cls.__name__
            new_type = type(base_cls_name, (cls, base_cls), {})
            register_object_type(new_type, vendor_id=842)
            instance, objectName, presentValue, description = args
            new_object = new_type(
                objectIdentifier=(base_cls.objectType, instance),
                objectName="{}".format(objectName),
                presentValue=presentValue,
                description=CharacterString("{}".format(description)),
            )
            return new_object

        return wrapper

    return decorate


def bacnet_property(property_name, value, *, force_mutable=None):
    """
    Given a property, add it to the object
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            allowed_prop = _allowed_prop(obj)
            mutable = _mutable(property_name)
            if property_name == "units":
                new_prop = EngineeringUnits.enumerations[value]
                obj.units = new_prop
            else:
                try:
                    new_prop = Property(
                        property_name,
                        allowed_prop[property_name],
                        default=value,
                        mutable=mutable,
                    )
                except KeyError:
                    raise ValueError(
                        "Invalid property ({}) for object".format(property_name)
                    )
                obj.add_property(new_prop)
            return obj

        return wrapper

    return decorate


def bacnet_properties(properties):
    """
    Given a dict of properties, add them to the object
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if callable(func):
                obj = func(*args, **kwargs)
            else:
                obj = func
            allowed_prop = _allowed_prop(obj)

            for property_name, value in properties.items():
                if property_name == "units":
                    new_prop = EngineeringUnits.enumerations[value]
                    obj.units = new_prop
                else:
                    try:
                        mutable = _mutable(property_name)
                        new_prop = Property(
                            property_name,
                            allowed_prop[property_name],
                            default=value,
                            mutable=mutable,
                        )
                    except KeyError:
                        raise ValueError(
                            "Invalid property ({}) for object".format(property_name)
                        )
                    obj.add_property(new_prop)
            return obj

        return wrapper

    return decorate


def create(object_type, instance, objectName, presentValue, description):
    new_object = object_type(
        objectIdentifier=(object_type.objectType, instance),
        objectName="{}".format(objectName),
        presentValue=presentValue,
        description=CharacterString("{}".format(description)),
    )
    return new_object



from bacpypes.object import Object, Property, register_object_type

# Prochaine étape : créer une focntion qui va lire "all" et se redéfinir dynamiquement
def create_proprietary_object(params):
    try:
        _validate_params(params)
    except:
        raise
    props = [
        Property(v["obj_id"], v["primitive"], mutable=v["mutable"])
        for k, v in params["properties"].items()
    ]

    new_class = type(
        params["name"],
        (params["bacpypes_type"],),
        {"objectType": params["objectType"], "properties": props},
    )
    register_object_type(new_class, vendor_id=params["vendor_id"])


def _validate_params(params):
    if not params["name"]:
        raise ValueError(
            "Proprietary definition dict must contains a name key with a custom class name"
        )
    if not params["vendor_id"]:
        raise ValueError("Vendor ID is mandatory")
    if not isinstance(params["properties"], dict):
        raise TypeError(
            "The definition must include a dict of properties. It can be empty."
        )
    if not issubclass(params["bacpypes_type"], Object):
        raise TypeError("bacpypes_type must be a subclass of bacpypes.object.Object")
    return True



#!/usr/bin/python
# -*- coding: utf-8 -*-

try:
    import bacpypes
except ImportError:
    # Using print here or setup.py will fail
    print("=" * 80)
    print(
        'bacpypes module missing, please install latest version using \n    $ "pip install bacpypes"'
    )
    print("\nDiscard this message if you are actually installing BAC0.")
    print("=" * 80)

try:
    from . import core
    from . import tasks
    from .scripts.Base import Base
    from .core.devices.Device import Device as device
    from .core.devices.Device import DeviceLoad as load
    from .core.devices.Trends import TrendLog as TrendLog
    from .tasks.Poll import SimplePoll as poll
    from .tasks.Match import Match as match
    from .tasks.Devices import AddDevice as add_device
    from .core.utils.notes import update_log_level as log_level
    from .infos import __version__ as version

    # To be able to use the complete version pandas, flask and bokeh must be installed.
    try:
        import pandas
        import bokeh
        import flask
        import flask_bootstrap

        _COMPLETE = True
    except ImportError:
        _COMPLETE = False

    from .scripts.Lite import Lite as lite

    if _COMPLETE:
        from .scripts.Complete import Complete as gui

        connect = gui
    else:
        connect = lite
        web = lambda: print(
            "All features not available to run BAC0.web(). Some modules are missing (flask, flask-bootstrap, bokeh, pandas). See docs for details. To start BAC0, use BAC0.lite()"
        )

    # Import proprietary classes
    from .core.proprietary_objects import jci

except ImportError as error:
    print("=" * 80)
    print(
        'Import Error, refer to documentation or reinstall using \n    $ "pip install BAC0"\n {}'.format(
            error
        )
    )
    print("\nDiscard this message if you are actually installing BAC0.")
    print("=" * 80)
    # Probably installing the app...



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
Utility function to retrieve a functionnal IP and a correct broadcast IP
address.
Goal : not use 255.255.255.255 as a broadcast IP address as it is not
accepted by every devices (>3.8.38.1 bacnet.jar of Tridium Jace for example)

"""
from bacpypes.pdu import Address

from ..io.IOExceptions import NetworkInterfaceException

import socket
import subprocess
import ipaddress
import sys
import re
from ...core.utils.notes import note_and_log


@note_and_log
class HostIP:
    """
    Special class to identify host IP informations
    """

    def __init__(self, port=47808):
        ip = self._findIPAddr()
        mask = self._findSubnetMask(ip)
        self._port = port
        self.interface = ipaddress.IPv4Interface("{}/{}".format(ip, mask))

    @property
    def ip_address_subnet(self):
        """
        IP Address/subnet
        """
        return "{}/{}".format(
            self.interface.ip.compressed, self.interface.exploded.split("/")[-1]
        )

    @property
    def ip_address(self):
        """
        IP Address/subnet
        """
        return "{}".format(self.interface.ip.compressed)

    @property
    def address(self):
        """
        IP Address using bacpypes Address format
        """
        port = ""
        if self._port:
            port = ":{}".format(self._port)
        return Address(
            "{}/{}{}".format(
                self.interface.ip.compressed,
                self.interface.exploded.split("/")[-1],
                port,
            )
        )

    @property
    def mask(self):
        """
        Subnet mask
        """
        return self.interface.exploded.split("/")[-1]

    @property
    def port(self):
        """
        IP Port used
        """
        return self._port

    def _findIPAddr(self):
        """
        Retrieve the IP address connected to internet... used as
        a default IP address when defining Script

        :returns: IP Adress as String
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(("google.com", 0))
            addr = s.getsockname()[0]
            s.close()
        except socket.error:
            raise NetworkInterfaceException(
                "Impossible to retrieve IP, please provide one manually"
            )
        return addr

    def _findSubnetMask(self, ip):
        """
        Retrieve the broadcast IP address connected to internet... used as
        a default IP address when defining Script

        :param ip: (str) optionnal IP address. If not provided, default to getIPAddr()
        
        :returns: broadcast IP Adress as String
        """
        try:
            import netifaces

            interfaces = netifaces.interfaces()
            for nic in interfaces:
                addresses = netifaces.ifaddresses(nic)
                try:
                    for address in addresses[netifaces.AF_INET]:
                        if address["addr"] == ip:
                            return address["netmask"]
                except KeyError:
                    pass

            return "255.255.255.255"
        except ImportError:
            self._log.warning(
                "Netifaces not installed on your system. BAC0 can't detect the subnet.\nPlease provide subnet for now, we'll consider 255.255.255.0 (/24).\nYou can install netifaces using 'pip install netifaces'."
            )
            return "255.255.255.0"


def validate_ip_address(ip):
    result = True
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        if not isinstance(ip, Address):
            raise ValueError("Provide Address as bacpypes.Address object")
        s.bind(ip.addrTuple)
    except OSError as error:
        result = False
    finally:
        s.close()
    return result



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
read_mixin.py - Add ReadProperty and ReadPropertyMultiple to a device 
"""
# --- standard Python modules ---

# --- 3rd party modules ---

# --- this application's modules ---
from ....tasks.Poll import DeviceNormalPoll, DeviceFastPoll
from ...io.IOExceptions import (
    ReadPropertyMultipleException,
    NoResponseFromController,
    SegmentationNotSupported,
    BufferOverflow,
)
from ..Points import NumericPoint, BooleanPoint, EnumPoint, StringPoint, OfflinePoint
from ..Trends import TrendLog

# ------------------------------------------------------------------------------


def retrieve_type(obj_list, point_type_key):
    for point_type, point_address in obj_list:
        if point_type_key in str(point_type):
            yield (point_type, point_address)


def to_float_if_possible(val):
    try:
        return float(val)
    except:
        return val


class ReadPropertyMultiple:
    """
    Handle ReadPropertyMultiple for a device
    """

    def _batches(self, request, points_per_request):
        """
        Generator for creating 'request batches'.  Each batch contains a maximum of "points_per_request" 
        points to read.
        :params: request a list of point_name as a list
        :params: (int) points_per_request
        :returns: (iter) list of point_name of size <= points_per_request
        """
        for i in range(0, len(request), points_per_request):
            yield request[i : i + points_per_request]

    def _rpm_request_by_name(self, point_list):
        """
        :param point_list: a list of point
        :returns: (tuple) read request for each points, points
        """
        points = []
        requests = []
        for each in point_list:
            str_list = []
            point = self._findPoint(each, force_read=False)
            points.append(point)

            str_list.append(
                " {} {} presentValue".format(
                    point.properties.type, point.properties.address
                )
            )
            rpm_param = "".join(str_list)
            requests.append(rpm_param)

        return (requests, points)

    def read_multiple(
        self,
        points_list,
        *,
        points_per_request=25,
        discover_request=(None, 6),
        force_single=False
    ):
        """
        Read points from a device using a ReadPropertyMultiple request.
        [ReadProperty requests are very slow in comparison].

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Requesting many points results big requests that need segmentation.  Aim to request
        just the 'right amount' so segmentation can be avoided.  Determining the 'right amount'
        is often trial-&-error.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        if not self.properties.pss["readPropertyMultiple"] or force_single:
            self._log.warning("Read property Multiple Not supported")
            self.read_single(
                points_list, points_per_request=1, discover_request=discover_request
            )
        else:
            if not self.properties.segmentation_supported:
                points_per_request = 1

            if discover_request[0]:
                values = []
                info_length = discover_request[1]
                big_request = discover_request[0]

                for request in self._batches(big_request, points_per_request):
                    try:
                        request = "{} {}".format(
                            self.properties.address, "".join(request)
                        )
                        self._log.debug("RPM_Request: {} ".format(request))
                        try:
                            val = self.properties.network.readMultiple(
                                request, vendor_id=self.properties.vendor_id
                            )
                        except SegmentationNotSupported:
                            raise

                        # print('val : ', val, len(val), type(val))
                        if val == None:
                            self.properties.segmentation_supported = False
                            raise SegmentationNotSupported

                    except KeyError as error:
                        raise Exception("Unknown point name : {}".format(error))

                    except SegmentationNotSupported as error:
                        self.properties.segmentation_supported = False
                        # self.read_multiple(points_list,points_per_request=1, discover_request=discover_request)
                        self._log.warning("Segmentation not supported")
                        self._log.warning("Request too big...will reduce it")
                        if points_per_request == 1:
                            raise
                        self.read_multiple(
                            points_list,
                            points_per_request=1,
                            discover_request=discover_request,
                        )

                    else:
                        for points_info in self._batches(val, info_length):
                            values.append(points_info)
                return values

            else:
                big_request = self._rpm_request_by_name(points_list)
                i = 0
                for request in self._batches(big_request[0], points_per_request):
                    try:
                        request = "{} {}".format(
                            self.properties.address, "".join(request)
                        )
                        val = self.properties.network.readMultiple(
                            request, vendor_id=self.properties.vendor_id
                        )

                    except SegmentationNotSupported as error:
                        self.properties.segmentation_supported = False
                        self.read_multiple(
                            points_list,
                            points_per_request=1,
                            discover_request=discover_request,
                        )

                    except KeyError as error:
                        raise Exception("Unknown point name : {}".format(error))

                    else:
                        points_values = zip(big_request[1][i : i + len(val)], val)
                        i += len(val)
                        for each in points_values:
                            each[0]._trend(each[1])

    def read_single(
        self, points_list, *, points_per_request=1, discover_request=(None, 4)
    ):
        if discover_request[0]:
            values = []
            info_length = discover_request[1]
            big_request = discover_request[0]

            for request in self._batches(big_request, points_per_request):
                try:
                    request = "{} {}".format(self.properties.address, "".join(request))
                    val = self.properties.network.read(
                        request, vendor_id=self.properties.vendor_id
                    )

                except KeyError as error:
                    raise Exception("Unknown point name : {}".format(error))

                # Save each value to history of each point
                for points_info in self._batches(val, info_length):
                    values.append(points_info)

            return values

        else:
            big_request = self._rpm_request_by_name(points_list)
            i = 0
            for request in self._batches(big_request[0], points_per_request):
                try:
                    request = "{} {}".format(self.properties.address, "".join(request))
                    val = self.properties.network.read(
                        request, vendor_id=self.properties.vendor_id
                    )
                    points_values = zip(big_request[1][i : i + len(val)], val)
                    i += len(val)
                    for each in points_values:
                        each[0]._trend(each[1])

                except KeyError as error:
                    raise Exception("Unknown point name : {}".format(error))

    def _discoverPoints(self, custom_object_list=None):
        if custom_object_list:
            objList = custom_object_list
        else:
            try:
                objList = self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    vendor_id=self.properties.vendor_id,
                )

            except NoResponseFromController:
                self._log.error(
                    "No object list available. Please provide a custom list using the object_list parameter"
                )
                objList = []

            except (SegmentationNotSupported, BufferOverflow):
                objList = []
                number_of_objects = self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    arr_index=0,
                    vendor_id=self.properties.vendor_id,
                )

                for i in range(1, number_of_objects + 1):
                    objList.append(
                        self.properties.network.read(
                            "{} device {} objectList".format(
                                self.properties.address, self.properties.device_id
                            ),
                            arr_index=i,
                            vendor_id=self.properties.vendor_id,
                        )
                    )

        points = []
        trendlogs = {}

        def _process_new_objects(
            obj_cls=NumericPoint, obj_type="analog", objList=None, points_per_request=5
        ):
            """
            Template to generate BAC0 points instances from information coming from the network.
            """
            request = []
            new_points = []
            if obj_type == "analog":
                prop_list = "objectName presentValue units description"
            elif obj_type == "binary":
                prop_list = (
                    "objectName presentValue inactiveText activeText description"
                )
            elif obj_type == "multi":
                prop_list = "objectName presentValue stateText description"
            elif obj_type == "characterstringValue":
                prop_list = "objectName presentValue"
            else:
                raise ValueError("Unsupported objectType")

            list_of_obj = retrieve_type(objList, obj_type)
            for points, address in list_of_obj:
                request.append("{} {} {} ".format(points, address, prop_list))

            def _find_propid_index(key):
                _prop_list = prop_list.split(" ")
                for i, each in enumerate(_prop_list):
                    if key == each:
                        return i
                raise KeyError("{} not part of property list".format(key))

            try:
                points_info = self.read_multiple(
                    "",
                    discover_request=(request, len(prop_list.split(" "))),
                    points_per_request=points_per_request,
                )
            except SegmentationNotSupported:
                raise
            # Process responses and create point
            i = 0
            for each in retrieve_type(objList, obj_type):
                point_type = str(each[0])
                point_address = str(each[1])
                point_infos = points_info[i]
                i += 1

                pointName = point_infos[_find_propid_index("objectName")]
                presentValue = point_infos[_find_propid_index("presentValue")]
                if obj_type == "analog":
                    presentValue = float(presentValue)
                elif obj_type == "multi":
                    presentValue = int(presentValue)
                try:
                    point_description = point_infos[_find_propid_index("description")]
                except KeyError:
                    point_description = ""
                try:
                    point_units_state = point_infos[_find_propid_index("units")]
                except KeyError:
                    try:
                        point_units_state = point_infos[_find_propid_index("stateText")]
                    except KeyError:
                        try:
                            _inactive = point_infos[_find_propid_index("inactiveText")]
                            _active = point_infos[_find_propid_index("activeText")]
                            point_units_state = (_inactive, _active)
                        except KeyError:
                            if obj_type == "binary":
                                point_units_state = ("OFF", "ON")
                            elif obj_type == "multi":
                                point_units_state = [""]
                            else:
                                point_units_state = None

                try:
                    new_points.append(
                        obj_cls(
                            pointType=point_type,
                            pointAddress=point_address,
                            pointName=pointName,
                            description=point_description,
                            presentValue=presentValue,
                            units_state=point_units_state,
                            device=self,
                            history_size=self.properties.history_size,
                        )
                    )
                except IndexError:
                    self._log.warning(
                        "There has been a problem defining {} points. It is sometimes due to busy network. Please retry the device creation".format(
                            obj_type
                        )
                    )
                    raise
            return new_points

        points.extend(
            _process_new_objects(
                obj_cls=NumericPoint, obj_type="analog", objList=objList
            )
        )
        points.extend(
            _process_new_objects(
                obj_cls=BooleanPoint, obj_type="binary", objList=objList
            )
        )
        points.extend(
            _process_new_objects(obj_cls=EnumPoint, obj_type="multi", objList=objList)
        )
        points.extend(
            _process_new_objects(
                obj_cls=StringPoint, obj_type="characterstringValue", objList=objList
            )
        )

        # TrendLogs
        for each in retrieve_type(objList, "trendLog"):
            point_address = str(each[1])
            tl = TrendLog(point_address, self, read_log_on_creation=False)
            ldop_type, ldop_addr = (
                tl.properties.log_device_object_property.objectIdentifier
            )
            ldop_prop = tl.properties.log_device_object_property.propertyIdentifier
            trendlogs["{}_{}_{}".format(ldop_type, ldop_addr, ldop_prop)] = (
                tl.properties.object_name,
                tl,
            )

        self._log.debug("RPM Mixin : %s | %s | %s", objList, points, trendlogs)
        self._log.info("Ready!")
        return (objList, points, trendlogs)

    def poll(self, command="start", *, delay=10):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        if delay < 10:
            self.properties.fast_polling = True
            _poll_cls = DeviceFastPoll
        else:
            self.properties.fast_polling = False
            _poll_cls = DeviceNormalPoll

        if str(command).lower() not in ["stop", "start", "0", "False"]:
            self._log.error(
                'Bad argument for function. Needs "stop", "start", "0" or "False" or provide keyword arg (command or delay)'
            )
            return

        if (
            str(command).lower() == "stop"
            or command == False
            or command == 0
            or delay == 0
        ):

            if isinstance(self._polling_task.task, DeviceNormalPoll) or isinstance(
                self._polling_task.task, DeviceFastPoll
            ):
                self._polling_task.task.stop()
                while self._polling_task.task.is_alive():
                    pass

                self._polling_task.task = None
                self._polling_task.running = False
                self._log.info("Polling stopped")

        elif self._polling_task.task is None:
            self._polling_task.task = _poll_cls(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info(
                "Polling started, values read every {} seconds".format(delay)
            )

        elif self._polling_task.running:
            self._polling_task.task.stop()
            while self._polling_task.task.is_alive():
                pass
            self._polling_task.running = False
            self._polling_task.task = _poll_cls(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info(
                "Polling started, every values read each {} seconds".format(delay)
            )

        else:
            raise RuntimeError("Stop polling before redefining it")


class ReadProperty:
    """
    Handle ReadProperty for a device
    """

    def _batches(self, request, points_per_request):
        """
        Generator for creating 'request batches'.  Each batch contains a maximum of "points_per_request"
        points to read.
        :params: request a list of point_name as a list
        :params: (int) points_per_request
        :returns: (iter) list of point_name of size <= points_per_request
        """
        for i in range(0, len(request), points_per_request):
            yield request[i : i + points_per_request]

    def _rpm_request_by_name(self, point_list):
        """
        :param point_list: a list of point
        :returns: (tuple) read request for each points, points
        """
        points = []
        requests = []
        for each in point_list:
            str_list = []
            point = self._findPoint(each, force_read=False)
            points.append(point)

            str_list.append(" " + point.properties.type)
            str_list.append(" " + str(point.properties.address))
            str_list.append(" presentValue")
            rpm_param = "".join(str_list)
            requests.append(rpm_param)

        return (requests, points)

    def read_multiple(
        self, points_list, *, points_per_request=1, discover_request=(None, 6)
    ):
        """
        Functions to read points from a device using the ReadPropertyMultiple request.
        Using readProperty request can be very slow to read a lot of data.

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Using too many points will create big requests needing segmentation.
        It's better to use just enough request so the message will not require
        segmentation.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        if isinstance(points_list, list):
            (requests, points) = self._rpm_request_by_name(points_list)
            for (i, req) in enumerate(requests):
                val = self.read_single(
                    req, points_per_request=1, discover_request=discover_request
                )
                if val is not None and val != "":
                    points[i]._trend(val)
        else:
            self.read_single(
                points_list, points_per_request=1, discover_request=discover_request
            )

    def read_single(self, request, *, points_per_request=1, discover_request=(None, 4)):
        try:
            request = "{} {}".format(self.properties.address, "".join(request))
            return self.properties.network.read(
                request, vendor_id=self.properties.vendor_id
            )
        except KeyError as error:
            raise Exception("Unknown point name: {}".format(error))

        except NoResponseFromController as error:
            return ""

    def _discoverPoints(self, custom_object_list=None):
        if custom_object_list:
            objList = custom_object_list
        else:
            try:
                objList = self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    vendor_id=self.properties.vendor_id,
                )

            except SegmentationNotSupported:
                objList = []
                number_of_objects = self.properties.network.read(
                    "{} device {} objectList".format(
                        self.properties.address, self.properties.device_id
                    ),
                    arr_index=0,
                    vendor_id=self.properties.vendor_id,
                )

                for i in range(1, number_of_objects + 1):
                    objList.append(
                        self.properties.network.read(
                            "{} device {} objectList".format(
                                self.properties.address, self.properties.device_id
                            ),
                            arr_index=i,
                            vendor_id=self.properties.vendor_id,
                        )
                    )

        points = []
        trendlogs = {}

        def _process_new_objects(obj_cls=NumericPoint, obj_type="analog", objList=None):
            _newpoints = []
            for each in retrieve_type(objList, obj_type):
                point_type = str(each[0])
                point_address = str(each[1])

                if obj_type == "analog":
                    units_state = self.read_single(
                        "{} {} units ".format(point_type, point_address)
                    )
                elif obj_type == "multi":
                    units_state = self.read_single(
                        "{} {} stateText ".format(point_type, point_address)
                    )
                elif obj_type == "binary":
                    units_state = (
                        (
                            self.read_single(
                                "{} {} inactiveText ".format(point_type, point_address)
                            )
                        ),
                        (
                            self.read_single(
                                "{} {} activeText ".format(point_type, point_address)
                            )
                        ),
                    )
                else:
                    units_state = None

                presentValue = self.read_single(
                    "{} {} presentValue ".format(point_type, point_address)
                )
                if obj_type == "analog":
                    presentValue = float(presentValue)

                _newpoints.append(
                    obj_cls(
                        pointType=point_type,
                        pointAddress=point_address,
                        pointName=self.read_single(
                            "{} {} objectName ".format(point_type, point_address)
                        ),
                        description=self.read_single(
                            "{} {} description ".format(point_type, point_address)
                        ),
                        presentValue=presentValue,
                        units_state=units_state,
                        device=self,
                    )
                )
            return _newpoints

        points.extend(_process_new_objects(NumericPoint, "analog", objList))
        points.extend(_process_new_objects(BooleanPoint, "binary", objList))
        points.extend(_process_new_objects(EnumPoint, "multi", objList))
        points.extend(
            _process_new_objects(StringPoint, "characterstringValue", objList)
        )

        for each in retrieve_type(objList, "trendLog"):
            point_address = str(each[1])
            try:
                tl = TrendLog(point_address, self)
            except Exception:
                self._log.error("Problem creating {}".format(each))
                continue
            ldop_type, ldop_addr = (
                tl.properties.log_device_object_property.objectIdentifier
            )
            ldop_prop = tl.properties.log_device_object_property.propertyIdentifier
            trendlogs["{}_{}_{}".format(ldop_type, ldop_addr, ldop_prop)] = (
                tl.properties.object_name,
                tl,
            )

        self._log.info("Ready!")
        return (objList, points, trendlogs)

    def poll(self, command="start", *, delay=120):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        if delay < 10:
            self._log.warning(
                "Device do not support RPM, fast polling not available, limiting delay to 10sec."
            )
            self.properties.fast_polling = False
            delay = 10

        if (
            str(command).lower() == "stop"
            or command == False
            or command == 0
            or delay == 0
        ):

            if isinstance(self._polling_task.task, DeviceNormalPoll):
                self._polling_task.task.stop()
                while self._polling_task.task.is_alive():
                    pass

                self._polling_task.task = None
                self._polling_task.running = False
                self._log.info("Polling stopped")

        elif self._polling_task.task is None:
            self._polling_task.task = DeviceNormalPoll(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info(
                "Polling started, values read every {} seconds".format(delay)
            )

        elif self._polling_task.running:
            self._polling_task.task.stop()
            while self._polling_task.task.is_alive():
                pass
            self._polling_task.running = False
            self._polling_task.task = DeviceNormalPoll(
                self, delay=delay, name=self.properties.name
            )
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info(
                "Polling started, every values read each {} seconds".format(delay)
            )

        else:
            raise RuntimeError("Stop polling before redefining it")



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Match.py - verify a point's status matches its commanded value.

Example:
    Is a fan commanded to 'On' actually 'running'?
"""

# --- standard Python modules ---
# --- 3rd party modules ---

# --- this application's modules ---
from .TaskManager import Task
from ..core.io.IOExceptions import BadDeviceDefinition
from ..core.devices.Device import Device
from ..core.utils.notes import note_and_log

"""
A way to define a BAC0.device using a task, so it won't block the Notebook or the REPL
"""


class AddDevice(Task):
    def __init__(self, address=None, boid=None, network=None, callback=None, **kwargs):
        if network is None:
            raise ValueError("Please provide network")
        else:
            self.network = network
        self._kwargs = kwargs

        if address is not None and boid is not None:
            self.address = address
            self.boid = boid
        else:
            self.address, self.boid = self.find_address(address, boid)

        name = "Adding_Device_{}|{}".format(self.address, self.boid)
        self.callback = callback
        self._kwargs["address"] = self.address
        self._kwargs["device_id"] = self.boid
        self._kwargs["network"] = network
        super().__init__(name=name, delay=0)
        # self.start()

    def find_address(self, address, boid):
        if self.network.discoveredDevices is None:
            self._log.error(
                "Device cannot be created yet, use bacnet.discover() or provide both address and Boid"
            )
            raise DeviceNotFoundError(
                "Device cannot be created yet, use bacnet.discover() or provide both address and Boid"
            )
        else:
            for each in self.network.discoveredDevices:
                address, boid = each
                if address in str(address) or boid in str(boid):
                    self._log.info("Found {}".format(each))
                    return each
        self._log.error(
            "Device not discovered yet, use bacnet.discover() or provide both address and Boid"
        )

    def task(self, **kwargs):
        try:
            self._log.info(kwargs)
            dev = Device(kwargs)
            self._log.info(
                "Device named {} ({}/{}) created. Retrieve it using bacnet[boid]".format(
                    dev.properties.name,
                    dev.properties.address,
                    dev.properties.device_id,
                )
            )
            if self.callback is not None:
                self._log.info("Executing callback for {}".format(dev.properties.name))
                self.callback()
        except BadDeviceDefinition as error:
            self._log.error("Bad device definition ({})".format(error))


class DeviceNotFoundError(Exception):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
TimeSync.py - creation of time synch requests

"""
from ...core.io.Read import find_reason
from ..io.IOExceptions import (
    SegmentationNotSupported,
    ReadPropertyException,
    ReadPropertyMultipleException,
    NoResponseFromController,
    ApplicationNotStarted,
)
from ...core.utils.notes import note_and_log

# --- standard Python modules ---
import datetime as dt

# --- 3rd party modules ---
from bacpypes.pdu import Address, GlobalBroadcast, LocalBroadcast
from bacpypes.primitivedata import Date, Time
from bacpypes.basetypes import DateTime
from bacpypes.apdu import TimeSynchronizationRequest, UTCTimeSynchronizationRequest
from bacpypes.iocb import IOCB
from bacpypes.core import deferred


def _build_datetime(UTC=False):
    if UTC:
        _d = dt.datetime.utcnow().date()
        _t = dt.datetime.utcnow().time()
        _date = Date(
            year=_d.year - 1900, month=_d.month, day=_d.day, day_of_week=_d.isoweekday()
        ).value
        _time = Time(
            hour=_t.hour,
            minute=_t.minute,
            second=_t.second,
            hundredth=int(_t.microsecond / 10000),
        ).value
    else:
        _date = Date().now().value
        _time = Time().now().value
    return DateTime(date=_date, time=_time)


@note_and_log
class TimeSync:
    """
    Mixin to support Time Synchronisation from BAC0 to other devices
    """

    def time_sync(self, destination=None, datetime=None, UTC=False):
        """
        Take local time and send it to devices. User can also provide
        a datetime value (constructed following bacpypes.basetypes.Datetime
        format).

        To create a DateTime ::

            from bacpypes.basetypes import DateTime
            from bacpypes.primitivedata import Date, Time

            # Create date and time
            _date = Date('2019-08-05')
            _time = Time('16:45')

            # Create Datetime
            _datetime = DateTime(date=_date.value, time=_time.value)

            # Pass this to the function
            bacnet.time_sync(datetime=_datetime)


        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if not datetime:
            _datetime = _build_datetime(UTC=UTC)
        elif isinstance(datetime, DateTime):
            _datetime = datetime
        else:
            raise ValueError(
                "Please provide valid DateTime in bacpypes.basetypes.DateTime format"
            )

        # build a request
        if UTC:
            request = UTCTimeSynchronizationRequest(time=_datetime)
        else:
            request = TimeSynchronizationRequest(time=_datetime)

        if destination:
            if destination.lower() == "global":
                request.pduDestination = GlobalBroadcast()
            elif destination.lower() == "local":
                request.pduDestination = LocalBroadcast()
            else:
                try:
                    request.pduDestination = Address(destination)
                except (TypeError, ValueError):
                    self._log.warning(
                        "Destination unrecognized ({}), setting local broadcast".format(
                            destination
                        )
                    )
                    request.pduDestination = LocalBroadcast()
        else:
            request.pduDestination = LocalBroadcast()

        self._log.debug("{:>12} {}".format("- request:", request))

        iocb = IOCB(request)  # make an IOCB

        # pass to the BACnet stack
        deferred(self.this_application.request_io, iocb)

        # Unconfirmed request...so wait until complete
        iocb.wait()  # Wait for BACnet response
        year, month, day, dow = _datetime.date
        year = year + 1900
        hour, minutes, sec, msec = _datetime.time
        d = dt.datetime(year, month, day, hour, minutes, sec, msec)
        self._log.info("Time Sync Request sent to network : {}".format(d.isoformat()))



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This will start the Bokeh Server
"""
from threading import Thread
import weakref
from bokeh.server.server import Server
from tornado.ioloop import IOLoop

from ..core.utils.notes import note_and_log


@note_and_log
class Bokeh_Worker(Thread):

    # Init thread running server
    def __init__(self, dev_app, trends_app, notes_app, serverIP, *, daemon=True):
        Thread.__init__(self, daemon=daemon)
        self._dev_app_ref = weakref.ref(dev_app)
        self._trends_app_ref = weakref.ref(trends_app)
        self._notes_app_ref = weakref.ref(notes_app)
        self.IP = serverIP
        self.exitFlag = False

    def run(self):
        self.process()

    def process(self):
        while not self.exitFlag:
            self.task()

    def startServer(self):
        self.server = Server(
            {
                "/devices": self._dev_app_ref(),
                "/trends": self._trends_app_ref(),
                "/notes": self._notes_app_ref(),
            },
            io_loop=IOLoop(),
            allow_websocket_origin=[
                "{}:8111".format(self.IP),
                "{}:5006".format(self.IP),
                "{}:8111".format("localhost"),
                "{}:5006".format("localhost"),
            ],
        )
        self.server.start()
        self.server.io_loop.start()

    def task(self):
        try:
            self.startServer()
        except Exception as err:
            self._log.warning("Bokeh server already running", err)
            self.exitFlag = True

    def stop(self):
        self.bokeh_server.stop()
        self.exitFlag = True

    def beforeStop(self):
        """
        Action done when closing thread
        """
        pass



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 21:37:02 2017

@author: CTremblay
"""


def create_sidebar(dash_class="", devices_class="", trends_class=""):
    sb = """
    <div class="wrapper">
    <div class="sidebar" data-background-color="white" data-active-color="danger">
    <!--
    Tip 1: you can change the color of the sidebar's background using: data-background-color="white | black"
    Tip 2: you can change the color of the active button using the data-active-color="primary | info | success | warning | danger"
    -->
    <div class="sidebar-wrapper">
      <div class="logo">
      <a href="https://github.com/ChristianTremblay/BAC0" class="simple-text">BAC0</a>
      </div>
      <ul class="nav">
        <li {}><a href="/">
               <i class="ti-panel"></i>
               <p>Dashboard</p>
               </a>
        </li>
        <li {}><a href="/trends">
               <i class="ti-panel"></i>
               <p>Trend</p>
               </a>
        </li>
        <li {}><a href="/dash_devices">
               <i class="ti-view-list-alt"></i>
               <p>Devices</p>
               </a>
        </li>
      </ul>
      </div>
    </div>
    """.format(
        dash_class, trends_class, devices_class
    )
    return sb


def update_notifications(log, new_msg):
    notif_log = log
    max_notifications = 5
    if new_msg:
        notif_log.insert(0, new_msg)
    if len(notif_log) > max_notifications:
        notif_log.pop()

    notif_list = '<ul class="dropdown-menu">'
    for notif in notif_log:
        notif_list += '<li><a href="#">' + notif + "</a></li>"
    if len(notif_log) == 0:
        notif_list += '<li><a href="#">Nothing yet</a></li>'
    notif_list += "</ul>"
    return notif_list


def create_card(
    icon="ti-server",
    title="title",
    data="None",
    id_data="generic_data",
    foot_icon="ti-reload",
    foot_data="None",
    id_foot_data="generic_foot_data",
):
    card = """
        <div class="col-lg-3 col-sm-6">
        <div class="card">
          <div class="content">
            <div class="row">
              <div class="col-xs-5">
                <div class="icon-big icon-success text-center">
                  <i class="{}"></i>
                </div>
              </div>
                <div class="col-xs-7">
                  <div class="numbers">
                    <p>{}</p>
                    <div id="{}">{}</div>
                  </div>
                </div>
              </div>
            <div class="footer">
            <hr />
            <div class="stats">
            <i class="{}"></i> <div id="{}">{}</div>
            </div>
            </div>
            </div>
        </div>
        </div>
        """.format(
        icon, title, id_data, data, foot_icon, id_foot_data, foot_data
    )
    return card



from bacpypes.apdu import SubscribeCOVRequest, SimpleAckPDU, RejectPDU, AbortPDU
from bacpypes.iocb import IOCB
from bacpypes.core import deferred
from bacpypes.pdu import Address
from bacpypes.object import get_object_class, get_datatype
from bacpypes.constructeddata import Array
from bacpypes.primitivedata import Tag, ObjectIdentifier, Unsigned

from BAC0.core.io.Read import cast_datatype_from_tag

"""

using cov, we build a "context" which is turned into a subscription being sent to 
the destination.

Once the IOCB is over, the callback attached to it will execute (subscription_acknowledged)
and we'll get the answer

"""


class SubscriptionContext:
    next_proc_id = 1

    def __init__(self, address, objectID, confirmed=None, lifetime=None, callback=None):
        self.address = address
        self.subscriberProcessIdentifier = self.next_proc_id
        self.next_proc_id += 1
        self.monitoredObjectIdentifier = objectID
        self.issueConfirmedNotifications = confirmed
        self.lifetime = lifetime
        self.callback = callback

    def cov_notification(self, apdu):
        # make a rash assumption that the property value is going to be
        # a single application encoded tag
        source = apdu.pduSource
        object_changed = apdu.monitoredObjectIdentifier

        elements = {
            "source": source,
            "object_changed": object_changed,
            "properties": {},
        }
        for element in apdu.listOfValues:
            prop_id = element.propertyIdentifier
            datatype = get_datatype(object_changed[0], prop_id)
            value = element.value

            if not datatype:
                value = cast_datatype_from_tag(
                    element.value, object_changed[0], prop_id
                )
            else:
                # special case for array parts, others are managed by cast_out
                if issubclass(datatype, Array) and (
                    element.propertyArrayIndex is not None
                ):
                    if element.propertyArrayIndex == 0:
                        value = element.value.cast_out(Unsigned)
                    else:
                        value = element.value.cast_out(datatype.subtype)
                else:
                    value = element.value.cast_out(datatype)

            elements["properties"][prop_id] = value

        return elements


class CoV:
    """
    Mixin to support COV registration
    """

    def send_cov_subscription(self, request):
        self._log.debug("Request : {}".format(request))
        iocb = IOCB(request)
        self._log.debug("IOCB : {}".format(iocb))

        iocb.add_callback(self.subscription_acknowledged)

        # pass to the BACnet stack
        deferred(self.this_application.request_io, iocb)

    def subscription_acknowledged(self, iocb):
        if iocb.ioResponse:
            self._log.info("Subscription success")

        if iocb.ioError:
            self._log.error("Subscription failed. {}".format(iocb.ioError))

    def cov(self, address, objectID, confirmed=True, lifetime=0, callback=None):
        address = Address(address)
        context = self._build_cov_context(
            address, objectID, confirmed=confirmed, lifetime=lifetime, callback=callback
        )
        request = self._build_cov_request(context)

        self.send_cov_subscription(request)

    def cancel_cov(self, address, objectID, callback=None):
        address = Address(address)
        context = self._build_cov_context(
            address, objectID, confirmed=None, lifetime=None, callback=callback
        )
        request = self._build_cov_request(context)

        self.send_cov_subscription(request)

    def _build_cov_context(
        self, address, objectID, confirmed=True, lifetime=None, callback=None
    ):
        context = SubscriptionContext(
            address=address,
            objectID=objectID,
            confirmed=confirmed,
            lifetime=lifetime,
            callback=callback,
        )
        self.subscription_contexts[context.subscriberProcessIdentifier] = context

        if "context_callback" not in self.subscription_contexts.keys():
            self.subscription_contexts["context_callback"] = self.context_callback
        return context

    def _build_cov_request(self, context):
        request = SubscribeCOVRequest(
            subscriberProcessIdentifier=context.subscriberProcessIdentifier,
            monitoredObjectIdentifier=context.monitoredObjectIdentifier,
        )
        request.pduDestination = context.address

        # optional parameters
        if context.issueConfirmedNotifications is not None:
            request.issueConfirmedNotifications = context.issueConfirmedNotifications
        if context.lifetime is not None:
            request.lifetime = context.lifetime

        return request

    # def context_callback(self, elements, callback=None):
    def context_callback(self, elements):
        self._log.info("Received COV Notification for {}".format(elements))
        # if callback:
        #    callback()
        for device in self.registered_devices:
            if str(device.properties.address) == str(elements["source"]):
                device[elements["object_changed"]].cov_registered = True
                for prop, value in elements["properties"].items():
                    if prop == "presentValue":
                        device[elements["object_changed"]]._trend(value)
                    else:
                        device[elements["object_changed"]].properties.bacnet_properties[
                            prop
                        ] = value
                break



from .decorator import bacnet_properties, make_commandable, create

from ...utils.notes import note_and_log
from ....scripts.Base import Base
from ...app.ScriptApplication import BAC0Application

from bacpypes.object import (
    AnalogInputObject,
    AnalogValueObject,
    BinaryValueObject,
    Property,
    register_object_type,
    registered_object_types,
    DatePatternValueObject,
    ReadableProperty,
    WritableProperty,
    OptionalProperty,
)
from bacpypes.basetypes import (
    EngineeringUnits,
    DateTime,
    PriorityArray,
    StatusFlags,
    Reliability,
    Polarity,
)

from collections import namedtuple
from colorama import Fore, Back, Style


@note_and_log
class ObjectFactory(object):
    """
    This is an exploration of a method to create local objects in BAC0 instance.

    First you need to know what bacpypes class of object you want to create ex. AnalogValueObject 
    This class must be imported from bacpypes ex. from bacpypes.object import AnalogValueObject

    You must also define supplemental properties for the object in a dict.
    ex. : 
        properties = {"outOfService" : False,
                "relinquishDefault" : 0,
                "units": "degreesCelsius",
                "highLimit": 98}

    Then you can use the factory to create your object
    ex. : 

        av0 = ObjectFactory(AnalogValueObject, 1, 'av0', )

    """

    instances = {}

    definition = namedtuple(
        "Definition",
        "name, objectType, instance, properties, description, presentValue, is_commandable, relinquish_default",
    )

    objects = {}
    # In the future... should think about a way to store relinquish default values because on a restart
    # those should be restored.

    def __init__(
        self,
        objectType,
        instance,
        objectName,
        properties=None,
        description="",
        presentValue=None,
        is_commandable=False,
        relinquish_default=None,
    ):
        self._properties = ObjectFactory.default_properties(
            objectType, properties, is_commandable, relinquish_default
        )
        pv_datatype = ObjectFactory.get_pv_datatype(objectType)

        if not isinstance(presentValue, pv_datatype):
            try:
                presentValue = pv_datatype(presentValue)
            except:
                raise ValueError("Wrong datatype provided for presentValue")

        @bacnet_properties(self._properties)
        @make_commandable()
        def _create_commandable(
            objectType, instance, objectName, presentValue, description
        ):
            return create(objectType, instance, objectName, presentValue, description)

        @bacnet_properties(self._properties)
        def _create(objectType, instance, objectName, presentValue, description):
            return create(objectType, instance, objectName, presentValue, description)

        objectName, instance = self.validate_name_and_instance(
            objectType, objectName, instance
        )

        if is_commandable:
            self.objects[objectName] = _create_commandable(
                objectType, instance, objectName, presentValue, description
            )
        else:
            self.objects[objectName] = _create(
                objectType, instance, objectName, presentValue, description
            )

    def validate_instance(self, objectType, instance):
        _warning = True
        try:
            _set = self.instances[objectType.__name__]
        except KeyError:
            _set = set()

        if not instance:
            instance = 0
            _warning = False

        if instance in _set:
            while instance in _set:
                instance += 1
            if _warning:
                self._log.warning(
                    "Instance alreaday taken, using {} instead".format(instance)
                )

        _set.add(instance)
        self.instances[objectType.__name__] = _set
        return instance

    def validate_name_and_instance(self, objectType, objectName, instance):
        name_must_be_changed = False
        if objectName in self.objects.keys():
            name_must_be_changed = True
        instance = self.validate_instance(objectType, instance)
        if name_must_be_changed:
            objectName = "{}-{}".format(objectName, instance)
            self._log.warning("Name already taken, using {} instead".format(objectName))

        return (objectName, instance)

    @classmethod
    def from_dict(cls, definition):
        return cls(
            objectType=definition["objectType"],
            instance=definition["instance"],
            objectName=definition["name"],
            properties=definition["properties"],
            description=definition["description"],
            presentValue=definition["presentValue"],
            is_commandable=definition["is_commandable"],
            relinquish_default=definition["relinquish_default"],
        )

    @staticmethod
    def properties_for(objectType):
        prop_list = {}
        for prop in objectType.properties:
            prop_list[prop.identifier] = {
                "datatype": prop.datatype,
                "optional": prop.optional,
                "mutable": prop.mutable,
                "default": prop.default,
            }
        return prop_list

    @staticmethod
    def get_pv_datatype(objectType):
        for prop in objectType.properties:
            if prop.identifier == "presentValue":
                return prop.datatype
        raise KeyError("Unknown")

    @staticmethod
    def clear_objects():
        ObjectFactory.objects = {}
        ObjectFactory.instances = {}

    def add_objects_to_application(self, app):
        if isinstance(app, Base):
            app = app.this_application
        if not isinstance(app, BAC0Application):
            raise TypeError("Provide BAC0Application object or BAC0 Base instance")
        for k, v in self.objects.items():
            try:
                app.add_object(v)
                self._log.info("Adding {} to application.".format(k))
            except RuntimeError:
                self._log.warning(
                    "There is already an object named {} in application.".format(k)
                )

    def __repr__(self):
        return "{}".format(self.objects)

    @staticmethod
    def default_properties(
        objectType, properties, is_commandable=False, relinquish_default=None
    ):
        _properties = properties or {}
        if "statusFlags" not in _properties.keys():
            _properties["statusFlags"] = [0, 0, 0, 0]
        if "reliability" not in _properties.keys():
            _properties["reliability"] = Reliability(0)
        if (
            "analog" in objectType.__name__.lower()
            and "units" not in _properties.keys()
        ):
            raise ValueError("Provide Engineering Units in properties")
        if is_commandable and (
            "input" in objectType.__name__.lower()
            or "output" in objectType.__name__.lower()
        ):
            _properties["outOfService"] = False
        if is_commandable:
            _properties["priorityArray"] = PriorityArray()
            _properties["relinquishDefault"] = ObjectFactory.relinquish_default_value(
                objectType, relinquish_default
            )
        return _properties

    @staticmethod
    def relinquish_default_value(objectType, value):
        pv_datatype = ObjectFactory.get_pv_datatype(objectType)
        return pv_datatype(value)

    @staticmethod
    def inspect(bacnet_object):
        """
        A fun little snippet to inspect the properties of a BACnet object.
        """
        objectType = bacnet_object.__name__
        _repr = Fore.YELLOW + "*" * 100
        _repr += "\n| " + Fore.CYAN + "{}".format(objectType)
        _repr += Fore.YELLOW + "\n|" + "*" * 99
        _repr += "\n"
        _repr += (
            "| "
            + Fore.WHITE
            + "Support COV : {}\n".format(bacnet_object._object_supports_cov)
        )
        _repr += Fore.YELLOW + "|" + "=" * 99
        _repr += "\n\x1b[33m| {:30} \x1b[33m| {:26} \x1b[33m| {:8} \x1b[33m| {:8} \x1b[33m| {:8}".format(
            "IDENTIFIER", "DATACLASS", "OPTIONAL", "MUTABLE", "DEFAULT"
        )
        _repr += "\n|" + "=" * 99
        for prop in bacnet_object.properties:
            i, d, o, m, df = prop.__dict__.values()
            df = Fore.WHITE + df if df else Fore.WHITE + "None"
            o = Fore.GREEN + "Yes" if o else Fore.RED + "No"
            m = Fore.GREEN + "Yes" if m else Fore.RED + "No"
            _repr += (
                "\n"
                + Fore.WHITE
                + "\x1b[33m| \x1b[37m{:30} \x1b[33m| \x1b[37m{:26} \x1b[33m| {:13} \x1b[33m| {:13} \x1b[33m| {:13}".format(
                    i, d.__name__, o, m, df
                )
            )
        return _repr



#!/usr/bin/env python

"""
Rebuilt Commandable
"""

from bacpypes.debugging import bacpypes_debugging, ModuleLogger
from bacpypes.consolelogging import ConfigArgumentParser

from bacpypes.core import run
from bacpypes.task import OneShotTask
from bacpypes.errors import ExecutionError

from bacpypes.primitivedata import (
    BitString,
    CharacterString,
    Date,
    Integer,
    Double,
    Enumerated,
    OctetString,
    Real,
    Time,
    Unsigned,
)
from bacpypes.basetypes import (
    BinaryPV,
    ChannelValue,
    DateTime,
    DoorValue,
    PriorityValue,
    PriorityArray,
)
from bacpypes.object import (
    Property,
    ReadableProperty,
    WritableProperty,
    register_object_type,
    AccessDoorObject,
    AnalogOutputObject,
    AnalogValueObject,
    BinaryOutputObject,
    BinaryValueObject,
    BitStringValueObject,
    CharacterStringValueObject,
    DateValueObject,
    DatePatternValueObject,
    DateTimePatternValueObject,
    DateTimeValueObject,
    IntegerValueObject,
    LargeAnalogValueObject,
    LightingOutputObject,
    MultiStateOutputObject,
    MultiStateValueObject,
    OctetStringValueObject,
    PositiveIntegerValueObject,
    TimeValueObject,
    TimePatternValueObject,
    ChannelObject,
)

from bacpypes.app import BIPSimpleApplication
from bacpypes.local.object import CurrentPropertyListMixIn
from bacpypes.local.device import LocalDeviceObject

# some debugging
_debug = 0
_log = ModuleLogger(globals())


#
#   Commandable
#


@bacpypes_debugging
def Commandable(
    datatype,
    presentValue="presentValue",
    priorityArray="priorityArray",
    relinquishDefault="relinquishDefault",
):
    if _debug:
        Commandable._debug("Commandable %r ...", datatype)

    class _Commando(object):

        properties = [
            WritableProperty(presentValue, datatype),
            ReadableProperty(priorityArray, PriorityArray),
            ReadableProperty(relinquishDefault, datatype),
        ]

        _pv_choice = None

        def __init__(self, **kwargs):
            super(_Commando, self).__init__(**kwargs)

            # build a default value in case one is needed
            default_value = datatype().value
            if issubclass(datatype, Enumerated):
                default_value = datatype._xlate_table[default_value]
            if _debug:
                Commandable._debug("    - default_value: %r", default_value)

            # see if a present value was provided
            if presentValue not in kwargs:
                setattr(self, presentValue, default_value)

            # see if a priority array was provided
            if priorityArray not in kwargs:
                setattr(self, priorityArray, PriorityArray())

            # see if a present value was provided
            if relinquishDefault not in kwargs:
                setattr(self, relinquishDefault, default_value)

        def _highest_priority_value(self):
            if _debug:
                Commandable._debug("_highest_priority_value")

            priority_array = getattr(self, priorityArray)
            for i in range(1, 17):
                priority_value = priority_array[i]
                if priority_value.null is None:
                    if _debug:
                        Commandable._debug("    - found at index: %r", i)

                    value = getattr(priority_value, _Commando._pv_choice)
                    value_source = "###"

                    if issubclass(datatype, Enumerated):
                        value = datatype._xlate_table[value]
                        if _debug:
                            Commandable._debug("    - remapped enumeration: %r", value)

                    break
            else:
                value = getattr(self, relinquishDefault)
                value_source = None

            if _debug:
                Commandable._debug(
                    "    - value, value_source: %r, %r", value, value_source
                )

            # return what you found
            return value, value_source

        def WriteProperty(
            self, property, value, arrayIndex=None, priority=None, direct=False
        ):
            if _debug:
                Commandable._debug(
                    "WriteProperty %r %r arrayIndex=%r priority=%r direct=%r",
                    property,
                    value,
                    arrayIndex,
                    priority,
                    direct,
                )

            # when writing to the presentValue with a priority
            if property == presentValue:
                if _debug:
                    Commandable._debug(
                        "    - writing to %s, priority %r", presentValue, priority
                    )

                # default (lowest) priority
                if priority is None:
                    priority = 16
                if _debug:
                    Commandable._debug(
                        "    - translate to priority array, index %d", priority
                    )

                # translate to updating the priority array
                property = priorityArray
                arrayIndex = priority
                priority = None

            # update the priority array entry
            if property == priorityArray:
                if arrayIndex is None:
                    if _debug:
                        Commandable._debug("    - writing entire %s", priorityArray)

                    # pass along the request
                    super(_Commando, self).WriteProperty(
                        property,
                        value,
                        arrayIndex=arrayIndex,
                        priority=priority,
                        direct=direct,
                    )
                else:
                    if _debug:
                        Commandable._debug(
                            "    - writing to %s, array index %d",
                            priorityArray,
                            arrayIndex,
                        )

                    # check the bounds
                    if arrayIndex == 0:
                        raise ExecutionError(
                            errorClass="property", errorCode="writeAccessDenied"
                        )
                    if (arrayIndex < 1) or (arrayIndex > 16):
                        raise ExecutionError(
                            errorClass="property", errorCode="invalidArrayIndex"
                        )

                    # update the specific priorty value element
                    priority_value = getattr(self, priorityArray)[arrayIndex]
                    if _debug:
                        Commandable._debug("    - priority_value: %r", priority_value)

                    # the null or the choice has to be set, the other clear
                    if value == ():
                        if _debug:
                            Commandable._debug("    - write a null")
                        priority_value.null = value
                        setattr(priority_value, _Commando._pv_choice, None)
                    else:
                        if _debug:
                            Commandable._debug("    - write a value")

                        if issubclass(datatype, Enumerated):
                            value = datatype._xlate_table[value]
                            if _debug:
                                Commandable._debug(
                                    "    - remapped enumeration: %r", value
                                )

                        priority_value.null = None
                        setattr(priority_value, _Commando._pv_choice, value)

                # look for the highest priority value
                value, value_source = self._highest_priority_value()

                # compare with the current value
                current_value = getattr(self, presentValue)
                if value == current_value:
                    if _debug:
                        Commandable._debug("    - no present value change")
                    return

                # turn this into a present value change
                property = presentValue
                arrayIndex = priority = None

            # allow the request to pass through
            if _debug:
                Commandable._debug(
                    "    - super: %r %r arrayIndex=%r priority=%r",
                    property,
                    value,
                    arrayIndex,
                    priority,
                )

            super(_Commando, self).WriteProperty(
                property, value, arrayIndex=arrayIndex, priority=priority, direct=direct
            )

    # look up a matching priority value choice
    for element in PriorityValue.choiceElements:
        if issubclass(datatype, element.klass):
            _Commando._pv_choice = element.name
            break
    else:
        _Commando._pv_choice = "constructedValue"
    if _debug:
        Commandable._debug("    - _pv_choice: %r", _Commando._pv_choice)

    # return the class
    return _Commando


#
#   MinOnOffTask
#


@bacpypes_debugging
class MinOnOffTask(OneShotTask):
    def __init__(self, binary_obj):
        if _debug:
            MinOnOffTask._debug("__init__ %s", repr(binary_obj))
        OneShotTask.__init__(self)

        # save a reference to the object
        self.binary_obj = binary_obj

        # listen for changes to the present value
        self.binary_obj._property_monitors["presentValue"].append(
            self.present_value_change
        )

    def present_value_change(self, old_value, new_value):
        if _debug:
            MinOnOffTask._debug("present_value_change %r %r", old_value, new_value)

        # if there's no value change, skip all this
        if old_value == new_value:
            if _debug:
                MinOnOffTask._debug("    - no state change")
            return

        # get the minimum on/off time
        if new_value == "inactive":
            task_delay = getattr(self.binary_obj, "minimumOnTime") or 0
            if _debug:
                MinOnOffTask._debug("    - minimum on: %r", task_delay)
        elif new_value == "active":
            task_delay = getattr(self.binary_obj, "minimumOffTime") or 0
            if _debug:
                MinOnOffTask._debug("    - minimum off: %r", task_delay)
        else:
            raise ValueError(
                "unrecognized present value for %r: %r"
                % (self.binary_obj.objectIdentifier, new_value)
            )

        # if there's no delay, don't bother
        if not task_delay:
            if _debug:
                MinOnOffTask._debug("    - no delay")
            return

        # set the value at priority 6
        self.binary_obj.WriteProperty("presentValue", new_value, priority=6)

        # install this to run, if there is a delay
        self.install_task(delta=task_delay)

    def process_task(self):
        if _debug:
            MinOnOffTask._debug("process_task(%s)", self.binary_obj.objectName)

        # clear the value at priority 6
        self.binary_obj.WriteProperty("presentValue", (), priority=6)


#
#   MinOnOff
#


@bacpypes_debugging
class MinOnOff(object):
    def __init__(self, **kwargs):
        if _debug:
            MinOnOff._debug("__init__ ...")
        super(MinOnOff, self).__init__(**kwargs)

        # create the timer task
        self._min_on_off_task = MinOnOffTask(self)


#
#   Commandable Standard Objects
#


class AccessDoorObjectCmd(Commandable(DoorValue), AccessDoorObject):
    pass


class AnalogOutputObjectCmd(Commandable(Real), AnalogOutputObject):
    pass


class AnalogValueObjectCmd(Commandable(Real), AnalogValueObject):
    pass


### class BinaryLightingOutputObjectCmd(Commandable(Real), BinaryLightingOutputObject):
###     pass


class BinaryOutputObjectCmd(Commandable(BinaryPV), MinOnOff, BinaryOutputObject):
    pass


class BinaryValueObjectCmd(Commandable(BinaryPV), MinOnOff, BinaryValueObject):
    pass


class BitStringValueObjectCmd(Commandable(BitString), BitStringValueObject):
    pass


class CharacterStringValueObjectCmd(
    Commandable(CharacterString), CharacterStringValueObject
):
    pass


class DateValueObjectCmd(Commandable(Date), DateValueObject):
    pass


class DatePatternValueObjectCmd(Commandable(Date), DatePatternValueObject):
    pass


class DateTimeValueObjectCmd(Commandable(DateTime), DateTimeValueObject):
    pass


class DateTimePatternValueObjectCmd(Commandable(DateTime), DateTimePatternValueObject):
    pass


class IntegerValueObjectCmd(Commandable(Integer), IntegerValueObject):
    pass


class LargeAnalogValueObjectCmd(Commandable(Double), LargeAnalogValueObject):
    pass


class LightingOutputObjectCmd(Commandable(Real), LightingOutputObject):
    pass


class MultiStateOutputObjectCmd(Commandable(Unsigned), MultiStateOutputObject):
    pass


class MultiStateValueObjectCmd(Commandable(Unsigned), MultiStateValueObject):
    pass


class OctetStringValueObjectCmd(Commandable(OctetString), OctetStringValueObject):
    pass


class PositiveIntegerValueObjectCmd(Commandable(Unsigned), PositiveIntegerValueObject):
    pass


class TimeValueObjectCmd(Commandable(Time), TimeValueObject):
    pass


class TimePatternValueObjectCmd(Commandable(Time), TimePatternValueObject):
    pass


#
#   ChannelValueProperty
#


class ChannelValueProperty(Property):
    def __init__(self):
        if _debug:
            ChannelValueProperty._debug("__init__")
        Property.__init__(
            self,
            "presentValue",
            ChannelValue,
            default=None,
            optional=False,
            mutable=True,
        )

    def WriteProperty(self, obj, value, arrayIndex=None, priority=None, direct=False):
        if _debug:
            ChannelValueProperty._debug(
                "WriteProperty %r %r arrayIndex=%r priority=%r direct=%r",
                obj,
                value,
                arrayIndex,
                priority,
                direct,
            )

        ### Clause 12.53.5, page 487
        raise NotImplementedError()


#
#   ChannelObjectCmd
#


class ChannelObjectCmd(ChannelObject):

    properties = [ChannelValueProperty()]


##
##
##
##
##


@register_object_type(vendor_id=999)
class LocalAnalogValueObjectCmd(CurrentPropertyListMixIn, AnalogValueObjectCmd):
    pass


@register_object_type(vendor_id=999)
class LocalBinaryOutputObjectCmd(CurrentPropertyListMixIn, BinaryOutputObjectCmd):
    pass


@register_object_type(vendor_id=999)
class LocalDateValueObjectCmd(CurrentPropertyListMixIn, DateValueObjectCmd):
    pass


"""
#
#   __main__
#

def main():
    # parse the command line arguments
    args = ConfigArgumentParser(description=__doc__).parse_args()

    if _debug: _log.debug("initialization")
    if _debug: _log.debug("    - args: %r", args)

    # make a device object
    this_device = LocalDeviceObject(ini=args.ini)
    if _debug: _log.debug("    - this_device: %r", this_device)

    # make a sample application
    this_application = BIPSimpleApplication(this_device, args.ini.address)

    # make a commandable analog value object, add to the device
    avo1 = LocalAnalogValueObjectCmd(
        objectIdentifier=('analogValue', 1),
        objectName='avo1',
        )
    if _debug: _log.debug("    - avo1: %r", avo1)
    this_application.add_object(avo1)

    # make a commandable binary output object, add to the device
    boo1 = LocalBinaryOutputObjectCmd(
        objectIdentifier=('binaryOutput', 1),
        objectName='boo1',
        presentValue='inactive',
        relinquishDefault='inactive',
        minimumOnTime=5,        # let it warm up
        minimumOffTime=10,      # let it cool off
        )
    if _debug: _log.debug("    - boo1: %r", boo1)
    this_application.add_object(boo1)

    # get the current date
    today = Date().now()

    # make a commandable date value object, add to the device
    dvo1 = LocalDateValueObjectCmd(
        objectIdentifier=('dateValue', 1),
        objectName='dvo1',
        presentValue=today.value,
        )
    if _debug: _log.debug("    - dvo1: %r", dvo1)
    this_application.add_object(dvo1)

    if _debug: _log.debug("running")

    run()

    _log.debug("fini")


if __name__ == "__main__":
    main()
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
SimpleApplication 
=================

A basic BACnet application (bacpypes BIPSimpleApplication) for interacting with 
the bacpypes BACnet stack.  It enables the base-level BACnet functionality 
(a.k.a. device discovery) - meaning it can send & receive WhoIs & IAm messages.

Additional functionality is enabled by inheriting this application, and then 
extending it with more functions. [See BAC0.scripts for more examples of this.]

"""
# --- standard Python modules ---
from collections import defaultdict

# --- 3rd party modules ---
from bacpypes.app import ApplicationIOController
from bacpypes.pdu import Address
from bacpypes.service.object import ReadWritePropertyMultipleServices
from bacpypes.service.cov import ChangeOfValueServices
from bacpypes.netservice import NetworkServiceAccessPoint, NetworkServiceElement
from bacpypes.bvllservice import (
    BIPSimple,
    BIPForeign,
    BIPBBMD,
    AnnexJCodec,
    UDPMultiplexer,
)
from bacpypes.apdu import SubscribeCOVRequest, SimpleAckPDU, RejectPDU, AbortPDU

from bacpypes.appservice import StateMachineAccessPoint, ApplicationServiceAccessPoint
from bacpypes.comm import ApplicationServiceElement, bind, Client
from bacpypes.iocb import IOCB
from bacpypes.core import deferred

# basic services
from bacpypes.service.device import WhoIsIAmServices, WhoHasIHaveServices
from bacpypes.service.object import ReadWritePropertyServices

# --- this application's modules ---
from ..utils.notes import note_and_log
from ..functions.Discover import NetworkServiceElementWithRequests

# ------------------------------------------------------------------------------


class common_mixin:
    """
    They take message coming from the network that are not generated from 
    a request we made.
    """

    def do_IAmRequest(self, apdu):
        """Given an I-Am request, cache it."""
        self._log.debug("do_IAmRequest {!r}".format(apdu))

        # build a key from the source, just use the instance number
        key = (str(apdu.pduSource), apdu.iAmDeviceIdentifier[1])
        self.i_am_counter[key] += 1
        self._last_i_am_received.append(key)

    def do_IHaveRequest(self, apdu):
        """Given an I-Have request, cache it."""
        self._log.debug("do_IHaveRequest {!r}".format(apdu))

        # build a key from the source, using object name
        key = (str(apdu.pduSource), apdu.objectName)
        self.i_have_counter[key] += 1
        self._last_i_have_received.append(key)

    def do_WhoIsRequest(self, apdu):
        """Respond to a Who-Is request."""

        # build a key from the source and parameters
        key = (
            str(apdu.pduSource),
            apdu.deviceInstanceRangeLowLimit,
            apdu.deviceInstanceRangeHighLimit,
        )
        self._log.debug(
            "do_WhoIsRequest from {} | {} to {}".format(key[0], key[1], key[2])
        )

        # count the times this has been received
        self.who_is_counter[key] += 1
        low_limit = key[1]
        high_limit = key[2]

        # count the times this has been received
        self.who_is_counter[key] += 1

        if low_limit is not None and self.localDevice.objectIdentifier[1] < low_limit:
            return
        if high_limit is not None and self.localDevice.objectIdentifier[1] > high_limit:
            return
        # generate an I-Am
        self._log.debug("Responding to Who is by a Iam")
        self.iam_req.pduDestination = apdu.pduSource
        iocb = IOCB(self.iam_req)  # make an IOCB
        deferred(self.request_io, iocb)

    def do_ConfirmedCOVNotificationRequest(self, apdu):
        # look up the process identifier
        context = self.subscription_contexts.get(apdu.subscriberProcessIdentifier, None)
        if not context or apdu.pduSource != context.address:
            self._log.warning(
                "Unsollicited COV Notification received from {} ({}). Have you restarted the application recently ?".format(
                    apdu.pduSource, apdu
                )
            )
            # this is turned into cancel_cov request and sent back to the client

        else:
            # now tell the context object
            elements = context.cov_notification(apdu)

            # success
            response = SimpleAckPDU(context=apdu)

            # send a confirmation
            self.response(response)
            self._log.debug("Confirmed COV Notification: {}".format(elements))
            self.subscription_contexts["context_callback"](elements)

            # execute callback
            if context.callback is not None:
                context.callback(elements=elements)

    def do_UnconfirmedCOVNotificationRequest(self, apdu):
        # look up the process identifier
        context = self.subscription_contexts.get(apdu.subscriberProcessIdentifier, None)
        if not context or apdu.pduSource != context.address:
            return

        # now tell the context object
        elements = context.cov_notification(apdu)
        self._log.debug("Unconfirmed COV Notification: {}".format(elements))
        self.subscription_contexts["context_callback"](elements)

        # execute callback
        if context.callback is not None:
            context.callback(elements=elements)


@note_and_log
class BAC0Application(
    common_mixin,
    ApplicationIOController,
    WhoIsIAmServices,
    WhoHasIHaveServices,
    ReadWritePropertyServices,
    ReadWritePropertyMultipleServices,
    ChangeOfValueServices,
):
    """
    Defines a basic BACnet/IP application to process BACnet requests.

    :param *args: local object device, local IP address
        See BAC0.scripts.BasicScript for more details.

    """

    def __init__(
        self,
        localDevice,
        localAddress,
        bbmdAddress=None,
        bbmdTTL=0,
        deviceInfoCache=None,
        aseID=None,
        iam_req=None,
        subscription_contexts=None,
    ):

        ApplicationIOController.__init__(
            self, localDevice, deviceInfoCache, aseID=aseID
        )

        self.iam_req = iam_req
        # local address might be useful for subclasses
        if isinstance(localAddress, Address):
            self.localAddress = localAddress
        else:
            self.localAddress = Address(localAddress)

        # include a application decoder
        self.asap = ApplicationServiceAccessPoint()

        # pass the device object to the state machine access point so it
        # can know if it should support segmentation
        self.smap = StateMachineAccessPoint(localDevice)

        # the segmentation state machines need access to the same device
        # information cache as the application
        self.smap.deviceInfoCache = self.deviceInfoCache

        # a network service access point will be needed
        self.nsap = NetworkServiceAccessPoint()

        # give the NSAP a generic network layer service element
        self.nse = NetworkServiceElementWithRequests()
        bind(self.nse, self.nsap)

        # bind the top layers
        bind(self, self.asap, self.smap, self.nsap)

        # create a generic BIP stack, bound to the Annex J server
        # on the UDP multiplexer
        self.bip = BIPSimple()
        self.annexj = AnnexJCodec()
        self.mux = UDPMultiplexer(self.localAddress)

        # bind the bottom layers
        bind(self.bip, self.annexj, self.mux.annexJ)

        # bind the BIP stack to the network, no network number
        self.nsap.bind(self.bip, address=self.localAddress)

        self.i_am_counter = defaultdict(int)
        self.i_have_counter = defaultdict(int)
        self.who_is_counter = defaultdict(int)

        # keep track of requests to line up responses
        self._request = None
        self._last_i_am_received = []
        self._last_i_have_received = []

        # to support CoV
        self.subscription_contexts = subscription_contexts

    def close_socket(self):
        # pass to the multiplexer, then down to the sockets
        self.mux.close_socket()

    def request(self, apdu):
        # save a copy of the request
        self._request = apdu

        # forward it along
        super(BAC0Application, self).request(apdu)


@note_and_log
class BAC0ForeignDeviceApplication(
    common_mixin,
    ApplicationIOController,
    WhoIsIAmServices,
    WhoHasIHaveServices,
    ReadWritePropertyServices,
    ReadWritePropertyMultipleServices,
    ChangeOfValueServices,
):
    """
    Defines a basic BACnet/IP application to process BACnet requests.

    :param *args: local object device, local IP address
        See BAC0.scripts.BasicScript for more details.

    """

    def __init__(
        self,
        localDevice,
        localAddress,
        bbmdAddress=None,
        bbmdTTL=0,
        deviceInfoCache=None,
        aseID=None,
        iam_req=None,
        subscription_contexts=None,
    ):

        ApplicationIOController.__init__(
            self, localDevice, deviceInfoCache, aseID=aseID
        )

        self.iam_req = iam_req
        # local address might be useful for subclasses
        if isinstance(localAddress, Address):
            self.localAddress = localAddress
        else:
            self.localAddress = Address(localAddress)

        # include a application decoder
        self.asap = ApplicationServiceAccessPoint()

        # pass the device object to the state machine access point so it
        # can know if it should support segmentation
        self.smap = StateMachineAccessPoint(localDevice)

        # the segmentation state machines need access to the same device
        # information cache as the application
        self.smap.deviceInfoCache = self.deviceInfoCache

        # a network service access point will be needed
        self.nsap = NetworkServiceAccessPoint()

        # give the NSAP a generic network layer service element
        self.nse = NetworkServiceElementWithRequests()
        bind(self.nse, self.nsap)

        # bind the top layers
        bind(self, self.asap, self.smap, self.nsap)

        # create a generic BIP stack, bound to the Annex J server
        # on the UDP multiplexer
        self.bip = BIPForeign(bbmdAddress, bbmdTTL)
        self.annexj = AnnexJCodec()
        self.mux = UDPMultiplexer(self.localAddress, noBroadcast=True)

        # bind the bottom layers
        bind(self.bip, self.annexj, self.mux.annexJ)

        # bind the NSAP to the stack, no network number
        self.nsap.bind(self.bip)

        self.i_am_counter = defaultdict(int)
        self.i_have_counter = defaultdict(int)
        self.who_is_counter = defaultdict(int)
        # keep track of requests to line up responses
        self._request = None
        self._last_i_am_received = []
        self._last_i_have_received = []

        # to support CoV
        self.subscription_contexts = subscription_contexts

    def close_socket(self):
        # pass to the multiplexer, then down to the sockets
        self.mux.close_socket()


class NullClient(Client):
    def __init__(self, cid=None):
        Client.__init__(self, cid=cid)

    def confirmation(self, *args, **kwargs):
        pass


@note_and_log
class BAC0BBMDDeviceApplication(
    common_mixin,
    ApplicationIOController,
    WhoIsIAmServices,
    WhoHasIHaveServices,
    ReadWritePropertyServices,
    ReadWritePropertyMultipleServices,
    ChangeOfValueServices,
):
    """
    Defines a basic BACnet/IP application to process BACnet requests.

    :param *args: local object device, local IP address
        See BAC0.scripts.BasicScript for more details.

    """

    bdt = []

    def __init__(
        self,
        localDevice,
        localAddress,
        bdtable=[],
        deviceInfoCache=None,
        aseID=None,
        iam_req=None,
        subscription_contexts=None,
    ):

        self.bdtable = bdtable

        null_client = NullClient()

        ApplicationIOController.__init__(
            self, localDevice, deviceInfoCache, aseID=aseID
        )

        self.iam_req = iam_req
        # local address might be useful for subclasses
        if isinstance(localAddress, Address):
            self.localAddress = localAddress
        else:
            self.localAddress = Address(localAddress)

        # include a application decoder
        self.asap = ApplicationServiceAccessPoint()

        # pass the device object to the state machine access point so it
        # can know if it should support segmentation
        self.smap = StateMachineAccessPoint(localDevice)

        # the segmentation state machines need access to the same device
        # information cache as the application
        self.smap.deviceInfoCache = self.deviceInfoCache

        # a network service access point will be needed
        self.nsap = NetworkServiceAccessPoint()

        # give the NSAP a generic network layer service element
        self.nse = NetworkServiceElementWithRequests()
        bind(self.nse, self.nsap)

        # bind the top layers
        bind(self, self.asap, self.smap, self.nsap)

        # create a generic BIP stack, bound to the Annex J server
        # on the UDP multiplexer
        self.bip = BIPBBMD(self.localAddress)
        self.annexj = AnnexJCodec()
        self.mux = UDPMultiplexer(self.localAddress, noBroadcast=True)

        # bind the bottom layers
        # bind(self.bip, self.annexj, self.mux.annexJ)
        bind(null_client, self.bip, self.annexj, self.mux.annexJ)

        if self.bdtable:
            for bdtentry in self.bdtable:
                self.add_peer(bdtentry)

        # bind the NSAP to the stack, no network number
        self.nsap.bind(self.bip)

        self.i_am_counter = defaultdict(int)
        self.i_have_counter = defaultdict(int)
        self.who_is_counter = defaultdict(int)
        # keep track of requests to line up responses
        self._request = None
        self._last_i_am_received = []
        self._last_i_have_received = []

        # to support CoV
        self.subscription_contexts = subscription_contexts

    def add_peer(self, address):
        try:
            bdt_address = Address(address)
            self.bip.add_peer(bdt_address)
        except Exception:
            raise

    def remove_peer(self, address):
        try:
            bdt_address = Address(address)
            self.bip.remove_peer(bdt_address)
        except Exception:
            raise

    def close_socket(self):
        # pass to the multiplexer, then down to the sockets
        self.mux.close_socket()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Reinitialize.py - creation of ReinitializeDeviceRequest

"""
# --- standard Python modules ---
import datetime as dt

# --- 3rd party modules ---
from bacpypes.pdu import Address, GlobalBroadcast
from bacpypes.primitivedata import Date, Time, CharacterString, Unsigned16
from bacpypes.basetypes import DateTime
from bacpypes.apdu import (
    DeviceCommunicationControlRequest,
    DeviceCommunicationControlRequestEnableDisable,
    SimpleAckPDU,
)
from bacpypes.iocb import IOCB
from bacpypes.core import deferred

from ...core.io.Read import find_reason
from ..io.IOExceptions import (
    SegmentationNotSupported,
    ReadPropertyException,
    ReadPropertyMultipleException,
    NoResponseFromController,
    ApplicationNotStarted,
)

from ...core.utils.notes import note_and_log


@note_and_log
class DeviceCommunicationControl:
    """
    Mixin to support DeviceCommunicationControl from BAC0 to other devices
    """

    def dcc(self, address=None, duration=None, password=None, state=None):
        """
        Will send DeviceCommunicationControl request
        """
        if not self._started:
            raise ApplicationNotStarted("BACnet stack not running - use startApp()")

        if not address:
            raise ValueError("Provide address for request")

        if not state:
            raise ValueError("Provide state ('enable', 'disable', 'disableInitiation'")

        # build a request
        request = DeviceCommunicationControlRequest()
        request.enableDisable = DeviceCommunicationControlRequestEnableDisable.enumerations[
            state
        ]
        request.pduDestination = Address(address)
        if duration:
            request.duration = Unsigned16(duration)

        request.password = CharacterString(password)

        self._log.debug("{:>12} {}".format("- request:", request))

        iocb = IOCB(request)  # make an IOCB

        # pass to the BACnet stack
        deferred(self.this_application.request_io, iocb)

        # Unconfirmed request...so wait until complete
        iocb.wait()  # Wait for BACnet response

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return

        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController("APDU Abort Reason : {}".format(reason))

        self._log.info(
            "DeviceCommunicationControl request sent to device : {}".format(address)
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Poll.py - create a Polling task to repeatedly read a point.
"""

# --- standard Python modules ---
import weakref

# --- 3rd party modules ---
from bacpypes.core import deferred

# --- this application's modules ---
from .TaskManager import Task
from ..core.utils.notes import note_and_log

# ------------------------------------------------------------------------------


@note_and_log
class SimplePoll(Task):
    """
    Start a polling task to repeatedly read a point's Present_Value.
    ex.
        device['point_name'].poll(delay=60)
    """

    def __init__(self, point, *, delay=10):
        """
        :param point: (BAC0.core.device.Points.Point) name of the point to read
        :param delay: (int) Delay between reads in seconds, defaults = 10sec
        
        A delay cannot be < 1sec
        This task is meant for single points, so BAC0 will allow short delays.
        This way, a fast polling is available for some points in a device that 
        would not support segmentation.

        :returns: Nothing
        """
        if delay < 1:
            delay = 1
        if point.properties:
            self._point = point
            Task.__init__(self, name="rp_poll", delay=delay)
        else:
            raise ValueError("Provide a point object")

    def task(self):
        self._point.value


@note_and_log
class DevicePoll(Task):
    """
    Start a polling task to repeatedly read a list of points from a device using 
    ReadPropertyMultiple requests.
    """

    def __init__(self, device, delay=10, name="", prefix="basic_poll"):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec
        
        A delay cannot be < 10sec 
        For delays under 10s, use DeviceFastPoll class.

        :returns: Nothing
        """
        self._device = weakref.ref(device)
        Task.__init__(self, name="{}_{}".format(prefix, name), delay=delay)
        self._counter = 0

    @property
    def device(self):
        return self._device()

    def task(self):
        try:
            self.device.read_multiple(
                list(self.device.points_name), points_per_request=25
            )
            self._counter += 1
            if self._counter == self.device.properties.auto_save:
                self.device.save(resampling=self.device.properties.save_resampling)
                if self.device.properties.clear_history_on_save:
                    self.device.clear_histories()
                self._counter = 0
        except AttributeError:
            # This error can be seen when defining a controller on a busy network...
            # When creation fail, polling is created and fail the first time...
            # So kill the task
            self.stop()
        except ValueError as e:
            self.device._log.error(
                "Something is wrong with polling...stopping. Try setting off "
                "segmentation. Error: {}".format(e)
            )
            self.stop()


@note_and_log
class DeviceNormalPoll(DevicePoll):
    """
    Start a normal polling task to repeatedly read a list of points from a device using 
    ReadPropertyMultiple requests.
    
    Normal polling will limit the polling speed to 10 second minimum

    """

    def __init__(self, device, delay=10, name=""):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec
        
        :returns: Nothing
        """
        if delay < 10:
            delay = 10
        self._log.info(
            "Device defined for normal polling with a delay of {}sec".format(delay)
        )
        DevicePoll.__init__(
            self, device=device, name=name, delay=delay, prefix="rpm_normal_poll"
        )


@note_and_log
class DeviceFastPoll(DevicePoll):
    """
    Start a fast polling task to repeatedly read a list of points from a device using 
    ReadPropertyMultiple requests.
    Delay allowed will be 0 to 10 seconds
    Normal polling will limit the polling speed to 10 second minimum

    Warning : Fast polling must be used with care or network flooding may occur

    """

    def __init__(self, device, delay=1, name=""):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 1sec
        
        :returns: Nothing
        """
        if delay < 0:
            delay = 0.01
        elif delay > 10:
            delay = 10
        self._log.warning(
            "Device defined for fast polling with a delay of {}sec".format(delay)
        )
        DevicePoll.__init__(
            self, device=device, name=name, delay=delay, prefix="rpm_fast_poll"
        )



from bacpypes.object import (
    MultiStateValueObject,
    AnalogValueObject,
    BinaryValueObject,
    AnalogInputObject,
    BinaryInputObject,
    AnalogOutputObject,
    BinaryOutputObject,
    CharacterStringValueObject,
    DateTimeValueObject,
    Property,
    register_object_type,
)

from bacpypes.local.object import (
    AnalogOutputCmdObject,
    AnalogValueCmdObject,
    BinaryOutputCmdObject,
    BinaryValueCmdObject,
)

from bacpypes.primitivedata import CharacterString, Date, Time, Real, Boolean
from bacpypes.constructeddata import ArrayOf
from bacpypes.basetypes import EngineeringUnits, DateTime, PriorityArray, StatusFlags

from .mixins.CommandableMixin import LocalBinaryOutputObjectCmd


def _make_mutable(obj, identifier="presentValue", mutable=True):
    """ 
    This function is not the way to go as it changes the class
    property...As bacpypes issue #224, it will need a lot of work
    """
    for prop in obj.properties:
        if prop.identifier == identifier:
            prop.mutable = mutable
    return obj


def create_MV(
    oid=1, pv=0, name="MV", states=["red", "green", "blue"], pv_writable=False
):
    msvo = MultiStateValueObject(
        objectIdentifier=("multiStateValue", oid),
        objectName=name,
        presentValue=pv,
        numberOfStates=len(states),
        stateText=ArrayOf(CharacterString)(states),
        priorityArray=PriorityArray(),
        statusFlags=StatusFlags(),
    )
    msvo = _make_mutable(msvo, mutable=pv_writable)
    deprecate_msg()
    return msvo


def create_AV(oid=1, pv=0, name="AV", units=None, pv_writable=False):
    avo = AnalogValueObject(
        objectIdentifier=("analogValue", oid),
        objectName=name,
        presentValue=pv,
        units=units,
        relinquishDefault=0,
        priorityArray=PriorityArray(),
        statusFlags=StatusFlags(),
    )
    avo = _make_mutable(avo, mutable=pv_writable)
    avo = _make_mutable(avo, identifier="relinquishDefault", mutable=pv_writable)
    deprecate_msg()
    return avo


def create_BV(
    oid=1, pv=0, name="BV", activeText="On", inactiveText="Off", pv_writable=False
):
    bvo = BinaryValueObject(
        objectIdentifier=("binaryValue", oid),
        objectName=name,
        presentValue=pv,
        activeText=activeText,
        inactiveText=inactiveText,
        priorityArray=PriorityArray(),
        statusFlags=StatusFlags(),
    )
    bvo = _make_mutable(bvo, mutable=pv_writable)
    deprecate_msg()
    return bvo


def create_AI(oid=1, pv=0, name="AI", units=None):
    aio = AnalogInputObject(
        objectIdentifier=("analogInput", oid),
        objectName=name,
        presentValue=pv,
        units=units,
        outOfService=Boolean(False),
        statusFlags=StatusFlags(),
    )
    aio = _make_mutable(aio, identifier="outOfService", mutable=True)
    deprecate_msg()
    return aio


def create_BI(oid=1, pv=0, name="BI", activeText="On", inactiveText="Off"):
    deprecate_msg()
    return BinaryInputObject(
        objectIdentifier=("binaryInput", oid),
        objectName=name,
        presentValue=pv,
        activeText=activeText,
        inactiveText=inactiveText,
        statusFlags=StatusFlags(),
    )


def create_AO(oid=1, pv=0, name="AO", units=None, pv_writable=False):
    aoo = AnalogOutputObject(
        objectIdentifier=("analogOutput", oid),
        objectName=name,
        presentValue=pv,
        units=units,
        priorityArray=PriorityArray(),
        statusFlags=StatusFlags(),
    )
    aoo = _make_mutable(aoo, mutable=pv_writable)
    deprecate_msg()
    return aoo


def create_BO(
    oid=1, pv=0, name="BO", activeText="On", inactiveText="Off", pv_writable=False
):
    boo = LocalBinaryOutputObjectCmd(
        objectIdentifier=("binaryOutput", oid),
        objectName=name,
        presentValue=pv,
        activeText=activeText,
        inactiveText=inactiveText,
        statusFlags=StatusFlags(),
    )
    boo = _make_mutable(boo, mutable=pv_writable)
    deprecate_msg()
    return boo


def create_CharStrValue(oid=1, pv="null", name="String", pv_writable=False):
    charval = CharacterStringValueObject(
        objectIdentifier=("characterstringValue", oid),
        objectName=name,
        priorityArray=PriorityArray(),
        statusFlags=StatusFlags(),
    )
    charval = _make_mutable(charval, mutable=pv_writable)
    charval.presentValue = CharacterString(pv)
    deprecate_msg()
    return charval


def create_DateTimeValue(
    oid=1, date=None, time=None, name="DateTime", pv_writable=False
):
    datetime = DateTimeValueObject(
        objectIdentifier=("datetimeValue", oid),
        objectName=name,
        statusFlags=StatusFlags(),
    )
    datetime = _make_mutable(datetime, mutable=pv_writable)
    datetime.presentValue = DateTime(date=Date(date), time=Time(time))
    deprecate_msg()
    return datetime


def create_object(
    object_class, oid, objectName, description, presentValue=None, commandable=False
):
    new_object = object_class(
        objectIdentifier=(object_class.objectType, oid),
        objectName="{}".format(objectName),
        presentValue=presentValue,
        description=CharacterString("{}".format(description)),
        statusFlags=StatusFlags(),
    )
    deprecate_msg()
    return _make_mutable(new_object, mutable=commandable)


def set_pv(obj=None, value=None, flags=[0, 0, 0, 0]):
    obj.presentValue = value
    obj.statusFlags = flags


def create_object_list(objects_dict):
    """
    d = {name: (name, description, presentValue, units, commandable)}
    """
    obj_list = []
    for obj_id, v in objects_dict.items():
        object_class, name, oid, description, presentValue, commandable = v
        description = CharacterString(description)
        new_obj = create_object(
            object_class, name, oid, description, commandable=commandable
        )
        if presentValue:
            new_obj.presentValue = presentValue
        obj_list.append(new_obj)
    return obj_list


def deprecate_msg():
    print("*" * 80)
    print("create_xx functions are deprecated and will disappear from a future release")
    print(
        "BAC0.core.device.local.object using the ObjectFactory will be the new way to define objects"
    )
    print("Refer to the doc for details")
    print("*" * 80)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from .jci import (
    JCIDeviceObject,
    JCIAnalogValueObject,
    JCIAnalogInputObject,
    JCIAnalogOutputObject,
)
from .object import create_proprietary_object

create_proprietary_object(JCIAnalogValueObject)
create_proprietary_object(JCIAnalogInputObject)
create_proprietary_object(JCIAnalogOutputObject)
create_proprietary_object(JCIDeviceObject)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This module deals with Bokeh Session, Document and Plots
A connection to the server is mandatory to use update_data
"""
from bokeh.plotting import Figure
from bokeh.models import (
    ColumnDataSource,
    HoverTool,
    Range1d,
    Label,
    LabelSet,
    LinearAxis,
    Legend,
    LegendItem,
    Dropdown,
    MultiChoice,
    CustomJS,
)
from bokeh.models.widgets import DataTable, TableColumn, Div
from bokeh.layouts import widgetbox, row, column, gridplot, widgetbox
from bokeh.palettes import d3, viridis
from bokeh.io import curdoc
from bokeh.application.handlers import Handler
from bokeh.models.tools import CustomJSHover

from functools import partial

from tornado import gen

import math
import pandas as pd
import weakref
from queue import Queue
from collections import deque

from ..tasks.RecurringTask import RecurringTask
from ..core.utils.notes import note_and_log
from ..core.devices.Virtuals import VirtualPoint
from ..core.devices.Trends import TrendLog


@note_and_log
class DynamicPlotHandler(Handler):

    analog_queue = Queue()
    binary_queue = Queue()
    multistates_queue = Queue()
    multistates_label_queue = Queue()
    virtuals_queue = Queue()

    def __init__(self, network):
        super().__init__()
        self._network = weakref.ref(network)

        self._last_time_list = None
        self._last_time_devices = None
        self._peristent_widget_choices = {}
        self._update_complete = False
        self._pcb = None  # Periodic callback
        self.widget = column(
            children=[MultiChoice(value=[], options=["n/a"])],
            sizing_mode="scale_width",
            name="widgetbox",
        )

        self._analog_name = ["A{}".format(each) for each in range(20)]
        self._binary_name = ["B{}".format(each) for each in range(20)]
        self._multistates_name = ["M{}".format(each) for each in range(20)]
        self._multistates_labels = ["M{}_state".format(each) for each in range(20)]
        self._virtuals_name = ["V{}".format(each) for each in range(20)]
        self.color_mappers = self.create_color_mappers()

        self._cds_struct = {}
        self.cds = self.initialize_cds()
        self.glyphs = {}

        self.figure = self.create_figure()

    @property
    def network(self):
        return self._network()

    @property
    def document(self):
        return curdoc()

    def clean_trend_list(self):
        a = []
        b = []
        m = []
        v = []
        for trend in self.network.trends:
            if "analog" in trend.properties.type:
                a.append(trend)
            elif "binary" in trend.properties.type:
                b.append(trend)
            elif "multi" in trend.properties.type:
                m.append(trend)
            else:
                v.append(trend)
        for each in a[19:]:
            self.warning(
                "Too many analog trends, removing {}".format(each.properties.name)
            )
            self.network.remove_trend(each)
        for each in b[19:]:
            self.warning(
                "Too many binary trends, removing {}".format(each.properties.name)
            )
            self.network.remove_trend(each)
        for each in m[19:]:
            self.warning(
                "Too many mutlistates trends, removing {}".format(each.properties.name)
            )
            self.network.remove_trend(each)
        for each in v[19:]:
            self.warning(
                "Too many virtual trends, removing {}".format(each.properties.name)
            )
            self.network.remove_trend(each)

    def build_dataframe(self):
        def _translate_binary_values(val):
            if val == "active:":
                return 0
            elif val == "inactive":
                return 1
            elif isinstance(val, str) and ":" in val:
                return int(val.split(":")[0])
            else:
                return val

        def _add_mv_states(val):
            try:
                return val.split(":")[1]
            except AttributeError:
                return val

        self._log.debug("Building dataframe")
        r = {}
        for point in self.network.trends:
            name = "{}/{}".format(
                point.properties.device.properties.name, point.properties.name
            )
            if point.history.dtype == "O":
                r[name] = point.history.apply(lambda x: _translate_binary_values(x))
            else:
                r[name] = point.history
            if "multi" in point.properties.type:
                # Create a series but keep only the first time the value change
                # Because I don't want to fill the trend with labels...
                _name = name + "_state"
                _duplicate = name + "_duplicate"
                r[_name] = point.history.apply(lambda x: _add_mv_states(x))
                r[_duplicate] = r[name].eq(r[name].shift())
                r[_name].loc[r[_duplicate]] = ""
                del r[_duplicate]

        try:
            df = pd.DataFrame(r)
        except ValueError:
            self._log.error("Problem with dataframe creation. {}".format(r.keys()))
            self.trouble_r = r
            raise
        df = df.reset_index()
        df["time_s"] = df["index"].apply(str)
        try:
            df = (
                df.fillna(method="ffill")
                .fillna(method="bfill")
                .replace(["inactive", "active"], [0, 1])
            )
        except TypeError:
            df = df.fillna(method="ffill").fillna(method="bfill")
        return df

    def make_glyph_invisible(self):
        for _, v in self.glyphs["analog"].items():
            v.visible = False
        for _, v in self.glyphs["binary"].items():
            v.visible = False
        for _, v in self.glyphs["multistates"].items():
            v.visible = False
        for _, v in self.glyphs["multistates_labels"].items():
            v.visible = False
        for _, v in self.glyphs["virtual"].items():
            v.visible = False

    def update_glyphs(self):
        self._log.debug("Updating Glyphs")
        enumerated_axis_needed = True
        for point in self.network.trends:
            name = "{}/{}".format(
                point.properties.device.properties.name, point.properties.name
            )
            if "analog" in point.properties.type or "TrendLog" in point.properties.type:
                self.analog_queue.put(
                    (name, point.properties.description, point.properties.units_state)
                )
            elif "binary" in point.properties.type:
                self.binary_queue.put(
                    (name, point.properties.description, point.properties.units_state)
                )
            elif "multi" in point.properties.type:
                enumerated_axis_needed = True
                self.multistates_queue.put(
                    (name, point.properties.description, point.properties.units_state)
                )
                self.multistates_label_queue.put(
                    (
                        name + "_state",
                        point.properties.description,
                        point.properties.units_state,
                    )
                )
            elif "virtual" in point.properties.type:
                self.virtuals_queue.put(
                    (name, point.properties.description, point.properties.units_state)
                )

        # for each in self.figure.right:
        #    try:
        #        if each.axis_label == "Enumerated":
        #            each.visible = enumerated_axis_needed
        #    except AttributeError:
        #        pass
        for each in [
            (self.analog_queue, "analog"),
            (self.binary_queue, "binary"),
            (self.multistates_queue, "multistates"),
            (self.multistates_label_queue, "multistates_labels"),
            (self.virtuals_queue, "virtual"),
        ]:
            i = 0
            process_queue, key = each

            while not process_queue.empty():

                index = "{}{}".format(key[0].upper(), i)
                if "labels" in key:
                    index = index + "_state"
                name, description, units = process_queue.get()
                self._log.debug("Processing {} on index {}".format(name, index))
                self._cds_struct[index] = name
                self.glyphs[key][index].name = name
                self.glyphs[key][index].tags = [units, description]
                self.glyphs[key][index].visible = True
                self._log.debug(
                    "{} | Visible : {}".format(
                        self.glyphs[key][index], self.glyphs[key][index].visible
                    )
                )
                i += 1
        self._log.debug("Update of Glyphs over")

    def initialize_cds(self):
        for each in range(20):
            analog_name = "A{}".format(each)
            binary_name = "B{}".format(each)
            multistates_name = "M{}".format(each)
            multistates_states = "M{}_state".format(each)
            virtuals_name = "V{}".format(each)
            _temp = {
                analog_name: None,
                binary_name: None,
                multistates_name: None,
                multistates_states: None,
                virtuals_name: None,
            }
            self._cds_struct.update(_temp)

        self._cds_struct.update({"index": "index", "time_s": "time_s"})

        df = self.build_dataframe()
        self.new_data = {}
        # add keys
        for k, v in self._cds_struct.items():
            self.new_data[k] = None

        for k, v in self.new_data.items():
            try:
                self.new_data[k] = df[self._cds_struct[k]].values
            except KeyError:
                self.new_data[k] = df["time_s"].values

        return ColumnDataSource(self.new_data)

    def update(self):
        df = self.build_dataframe()

        self.new_data = {}
        # add keys
        for k, v in self._cds_struct.items():
            self.new_data[k] = None

        for k, v in self.new_data.items():
            try:
                self.new_data[k] = df[self._cds_struct[k]].values

            except (KeyError, ValueError):
                self.new_data[k] = df["time_s"].values

        try:
            self.cds.data = self.new_data
        except RuntimeError as error:
            self._log.warning("Update failed, will try later | {}".format(error))

        # add selected points to bacnet.trend
        def get_device(name):
            for controller in self.network.registered_devices:
                if controller.properties.name == name:
                    return controller
            raise KeyError("Controller not found")

        def _already_trended(name):
            # print('Point name : {}'.format(name))
            for point in self.network.trends:
                if point.properties.name == name:
                    return True
            return False

        self._all_points_selected = []

        for device_choice in self.widget.children:
            try:
                _device = get_device(device_choice.title)
            except KeyError:
                continue
            self._peristent_widget_choices[device_choice.title] = device_choice.value
            for point_name in device_choice.value:
                self._all_points_selected.append(
                    "{}/{}".format(device_choice.title, point_name)
                )
                if not _already_trended(point_name):
                    _name = "{}/{}".format(device_choice.title, point_name)
                    self._log.info("Adding {} to trends".format(_name))
                    _device[point_name].chart()

        #       self._all_points_selected = [device_choice.value for device_choice in self.widget.children][0]
        for point in self.network.trends:
            _trend_name = "{}/{}".format(
                point.properties.device.properties.name, point.properties.name
            )
            if _trend_name not in self._all_points_selected and not isinstance(
                point, VirtualPoint
            ):
                self._log.info("Removing {} from trends".format(_trend_name))
                self.network.remove_trend(point)
                for k, v in self._cds_struct.items():
                    if v == _trend_name:
                        self._cds_struct[k] = None

                for device_choice in self.widget.children:
                    self._peristent_widget_choices[
                        device_choice.title
                    ] = device_choice.value

        if self.verify_if_document_needs_to_be_modified():
            try:
                self.update_glyphs()
                self.update_legend()
                self.update_widgets()
            except RuntimeError:
                self._log.warning("Problem updating. Will retry")

    def create_figure(self):

        analog_glyphs = {}
        binary_glyphs = {}
        multistates_glyphs = {}
        multistates_labels = {}
        virtuals_glyphs = {}

        self._log.debug("Creating figure")

        TOOLS = "pan,box_zoom,wheel_zoom,save,reset"
        p = Figure(
            x_axis_type="datetime",
            x_axis_label="Time",
            y_axis_label="Value",
            title="Live trends",
            tools=TOOLS,
            plot_width=1200,
            plot_height=800,
            toolbar_location="right",
        )

        p.title.text_font_size = "24pt"
        p.xaxis.axis_label_text_font_size = "18pt"
        p.yaxis.axis_label_text_font_size = "18pt"
        p.xaxis.axis_label_text_font_style = "normal"
        p.yaxis.axis_label_text_font_style = "normal"
        p.xaxis.major_label_text_font_size = "12pt"
        p.yaxis.major_label_text_font_size = "12pt"

        p.background_fill_color = "#f4f3ef"
        p.border_fill_color = "#f4f3ef"
        p.extra_y_ranges = {
            "bool": Range1d(start=0, end=1.1),
            "enum": Range1d(start=0, end=10),
        }
        p.add_layout(
            LinearAxis(y_range_name="bool", axis_label="Binary", visible=False), "left"
        )
        p.add_layout(
            LinearAxis(
                y_range_name="enum",
                axis_label="Enumerated",
                visible=False,
                # ticker=list(range(11)),
            ),
            "right",
        )

        hover_common = HoverTool(
            tooltips=[
                ("name", "$name"),
                ("value", "$data_y"),
                # ('state', "@$name_state"),
                # ("units", "$tags"),
                ("time", "@time_s"),
            ],
            renderers=[],
            toggleable=False,
            formatters={"@time_s": "datetime"},
            mode="mouse",
        )
        hover_multi = {}

        p.add_tools(hover_common)

        for name in self._binary_name:
            binary_glyphs[name] = p.step(
                x="index",
                y=name,
                source=self.cds,
                name=name,
                color=self.color_mappers["binary"][name],
                y_range_name="bool",
                mode="after",
                line_width=8,
                visible=False,
                tags=["unit", "description"],
            )
            # binary_glyphs[name].add_tool(hover_common)
            hover_common.renderers.append(binary_glyphs[name])

        for name in self._multistates_name:
            multistates_glyphs[name] = p.step(
                x="index",
                y=name,
                source=self.cds,
                name=name,
                color=self.color_mappers["multistates"][name],
                y_range_name="enum",
                line_dash="dashed",
                line_width=7,
                visible=False,
                tags=["unit", "description"],
                mode="after",
            )

        #        for name in self._multistates_labels:
        #            multistates_labels[name] = LabelSet(x="index", y=name.split('_')[0], text=name, level='glyph',
        #              x_offset=0, y_offset=1, source=self.cds, render_mode='canvas', visible=False)
        #            p.add_layout(multistates_labels[name])
        #        for name in self._multistates_labels:
        #            _msname = name.split("_")[0]
        #            multistates_labels[name] = p.circle(
        #                x="index",
        #                y=_msname,
        #                source=self.cds,
        #                color=self.color_mappers["multistates"][_msname],
        #                size=10,
        #                alpha=0.1,
        #                y_range_name="enum",
        #                visible=False,
        #            )
        #            hover_multi[name] = HoverTool(
        #                tooltips=[
        #                    ("name", "$name"),
        #                    ("value", "@" + name),
        #                    ("time", "@time_s"),
        #                ],
        #                mode="mouse",
        #                renderers=[multistates_labels[name]],
        #                toggleable=False,
        #            )
        #            p.add_tools(hover_multi[name])

        for name in self._multistates_labels:
            _msname = name.split("_")[0]
            multistates_labels[name] = p.text(
                x="index",
                y=_msname,
                text=name,
                source=self.cds,
                text_color=self.color_mappers["multistates"][_msname],
                angle=0.7835,
                # size=10,
                # alpha=0.1,
                y_range_name="enum",
                visible=False,
            )
            hover_multi[name] = HoverTool(
                tooltips=[
                    ("name", "$name"),
                    ("value", "@" + name),
                    ("time", "@time_s"),
                ],
                mode="mouse",
                renderers=[multistates_labels[name]],
                toggleable=False,
            )
            p.add_tools(hover_multi[name])

        for name in self._analog_name:
            analog_glyphs[name] = p.line(
                x="index",
                y=name,
                source=self.cds,
                name=name,
                color=self.color_mappers["analog"][name],
                line_width=2,
                visible=False,
                tags=["unit", "description"],
            )
            # analog_glyphs[name].add_tool(hover_common)
            hover_common.renderers.append(analog_glyphs[name])

        for name in self._virtuals_name:
            virtuals_glyphs[name] = p.line(
                x="index",
                y=name,
                source=self.cds,
                name=name,
                color=self.color_mappers["virtual"][name],
                line_width=2,
                visible=False,
                tags=["unit", "description"],
            )
            # virtuals_glyphs[name].add_tool(hover_common)
            hover_common.renderers.append(virtuals_glyphs[name])

        self.glyphs = {
            "analog": analog_glyphs,
            "binary": binary_glyphs,
            "multistates": multistates_glyphs,
            "multistates_labels": multistates_labels,
            "virtual": virtuals_glyphs,
        }
        legend = Legend(items=[])
        legend.click_policy = "hide"
        p.add_layout(legend, "below")
        return p

    @staticmethod
    def _type(k):
        try:
            t = {"A": "analog", "B": "binary", "M": "multistates", "V": "virtual"}
            _k = t[k[0]]
        except KeyError:
            _k = k
        return _k

    def update_legend(self):
        lst_of_items = []
        for k, v in self._cds_struct.items():
            if not v or "_state" in k:
                continue
            try:
                _t = DynamicPlotHandler._type(k)
                glyph = self.glyphs[_t][k]
                label = "{} | {} ({})".format(glyph.name, glyph.tags[1], glyph.tags[0])
                if _t != "M":
                    lst_of_items.append(LegendItem(label=label, renderers=[glyph]))
                else:
                    glyph_text = self.glyphs["multistates_labels"][k]
                    lst_of_items.append(
                        LegendItem(label=label, renderers=[glyph, glyph_text])
                    )
            except KeyError:
                pass

        self.figure.legend.items = lst_of_items

    def create_color_mappers(self):
        def rotate_palette(palette):
            for each in range(4):
                x = palette.pop()
                palette.insert(0, x)
            return palette

        palette = []
        _p = [d3["Category20"][20][i::4] for i in range(4)]
        for each in _p:
            palette.extend(each)
        keys = [each for each in self._analog_name]
        analog_color_mapper = dict(zip(keys, palette))

        palette = rotate_palette(palette)
        keys = [each for each in self._binary_name]
        binary_color_mapper = dict(zip(keys, palette))

        palette = rotate_palette(palette)
        keys = [each for each in self._multistates_name]
        multistates_color_mapper = dict(zip(keys, palette))

        palette = rotate_palette(palette)
        keys = [each for each in self._virtuals_name]
        virtual_color_mapper = dict(zip(keys, palette))

        return {
            "analog": analog_color_mapper,
            "binary": binary_color_mapper,
            "multistates": multistates_color_mapper,
            "virtual": virtual_color_mapper,
        }

    def modify_document(self, doc):
        self._log.debug("Modify document")
        self.stop_update_data()
        doc.clear()
        self._cds_struct = {}
        self.cds = self.initialize_cds()
        self.figure = self.create_figure()
        layout = row(
            [
                column([self.figure], name="main_plot"),
                column([self.widget], name="widgets"),
            ]
        )
        doc.add_root(layout)
        self._pcb = doc.add_periodic_callback(self.update, 3000)
        return doc

    def verify_if_document_needs_to_be_modified(self):
        # self.clean_trend_list()
        if (
            self._last_time_list != self.network.trends
            or self._last_time_devices != self.network.registered_devices
        ):
            self._log.debug(
                "List of points to trend have changed, document will be modified."
            )
            self._last_time_list = self.network.trends
            self._last_time_devices = self.network.registered_devices
            return True
        else:
            return False

    def stop_update_data(self):
        doc = curdoc()
        try:
            doc.remove_periodic_callback(self._pcb)
        except:
            pass

    def update_widgets(self):
        sel = self.generate_selection()
        self.widget.children = sel

    def generate_selection(self):
        devices = self.network.registered_devices
        selectors = []
        _cache = []

        for device in devices:
            try:
                _cache = self._peristent_widget_choices[device.properties.name]
            except KeyError:
                pass
            options = list(device.points_name)
            options.extend(list(device.trendlogs_names))
            mc = MultiChoice(
                value=_cache, options=options, title=device.properties.name
            )
            mc.js_on_change(
                "value",
                CustomJS(
                    code="""
                console.log('multi_choice: value=' + this.value, this.toString())
                """
                ),
            )

            selectors.append(mc)
        return selectors


class DevicesTableHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network()

    def modify_document(self, doc):
        self.network.whois()
        devices_df = self.network.devices
        dev = ColumnDataSource(devices_df)
        columns = [
            TableColumn(field=" Device ID", title="Dev ID"),
            TableColumn(field="Address", title="Address"),
            TableColumn(field="Manufacturer", title="Manuf"),
            TableColumn(field="Name", title="Name"),
        ]
        data_table = DataTable(source=dev, columns=columns)
        layout = row([data_table])
        doc.add_root(layout)
        doc.title = "BACnet devices"
        return doc


class NetworkPieChartHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network()

    def modify_document(self, doc):
        self.network.whois()
        devices_df = self.network.devices
        dev = ColumnDataSource(devices_df)
        columns = [
            TableColumn(field=" Device ID", title="Dev ID"),
            TableColumn(field="Address", title="Address"),
            TableColumn(field="Manufacturer", title="Manuf"),
            TableColumn(field="Name", title="Name"),
        ]
        data_table = DataTable(source=dev, columns=columns)
        layout = row([data_table])
        doc.add_root(layout)
        doc.title = "BACnet devices"
        return doc


class NotesTableHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network_ref = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network_ref()

    def modify_document(self, doc):

        controller = self.network.notes[0]
        notes_df = pd.DataFrame(self.network.notes[1]).reset_index()
        notes_df.columns = ["index", "notes"]
        notes = ColumnDataSource(notes_df)
        self.columns = [
            TableColumn(field="index", title="Timestamp"),
            TableColumn(field="notes", title="Notes"),
        ]
        self.data_table = DataTable(source=notes, columns=self.columns)
        layout = row([self.data_table])
        doc.add_root(layout)
        doc.title = "Notes for {}".format(controller)
        # doc.add_periodic_callback(self.update_data,100)
        return doc

    def update_data(self):
        controller = self.network.notes[0]
        notes_df = pd.DataFrame(self.network.notes[1]).reset_index()
        notes_df.columns = ["index", "notes"]
        notes = ColumnDataSource(notes_df)
        self.data_table.source.data.update(notes.data)
        curdoc().title = "Notes for {}".format(controller)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
BasicScript - implement the BAC0.core.app.ScriptApplication
Its basic function is to start and stop the bacpypes stack.
Stopping the stack, frees the IP socket used for BACnet communications.
No communications will occur if the stack is stopped.

Bacpypes stack enables Whois and Iam functions, since this minimum is needed to be
a BACnet device.  Other stack services can be enabled later (via class inheritance).
[see: see BAC0.scripts.ReadWriteScript]

Class::
    BasicScript(WhoisIAm)
        def startApp()
        def stopApp()
"""
# --- standard Python modules ---
from threading import Thread
import random
import sys

# --- 3rd party modules ---

from bacpypes.core import run as startBacnetIPApp
from bacpypes.core import stop as stopBacnetIPApp
from bacpypes.core import enable_sleeping
from bacpypes.local.device import LocalDeviceObject
from bacpypes.basetypes import ServicesSupported, DeviceStatus
from bacpypes.primitivedata import CharacterString

# --- this application's modules ---
from ..core.app.ScriptApplication import (
    BAC0Application,
    BAC0ForeignDeviceApplication,
    BAC0BBMDDeviceApplication,
)
from .. import infos
from ..core.io.IOExceptions import InitializationError, UnknownObjectError
from ..core.functions.GetIPAddr import validate_ip_address
from ..tasks.TaskManager import stopAllTasks

from ..core.utils.notes import note_and_log

try:
    import pandas
    import bokeh
    import flask
    import flask_bootstrap

    _COMPLETE = True
except ImportError:
    _COMPLETE = False

# ------------------------------------------------------------------------------


@note_and_log
class LocalObjects(object):
    def __init__(self, device):
        self.device = device

    def __getitem__(self, obj):
        item = None
        if isinstance(obj, tuple):
            obj_type, instance = obj
            item = self.device.this_application.get_object_id((obj_type, instance))
        elif isinstance(obj, str):
            name = obj
            item = self.device.this_application.get_object_name(name)
        if item is None:
            raise UnknownObjectError("Can't find {} in local device".format(obj))
        else:
            return item


@note_and_log
class Base:
    """
    Build a running BACnet/IP device that accepts WhoIs and IAm requests
    Initialization requires some minimial information about the local device.

    :param localIPAddr='127.0.0.1':
    :param localObjName='BAC0':
    :param deviceId=None:
    :param maxAPDULengthAccepted='1024':
    :param maxSegmentsAccepted='1024':
    :param segmentationSupported='segmentedBoth':
    """

    _used_ips = set()

    def __init__(
        self,
        localIPAddr="127.0.0.1",
        localObjName="BAC0",
        deviceId=None,
        firmwareRevision="".join(sys.version.split("|")[:2]),
        maxAPDULengthAccepted="1024",
        maxSegmentsAccepted="1024",
        segmentationSupported="segmentedBoth",
        bbmdAddress=None,
        bbmdTTL=0,
        bdtable=None,
        modelName=CharacterString("BAC0 Scripting Tool"),
        vendorId=842,
        vendorName=CharacterString("SERVISYS inc."),
        description=CharacterString("http://christiantremblay.github.io/BAC0/"),
    ):

        self._log.debug("Configurating app")

        if not _COMPLETE:
            self._log.debug(
                "To be able to run the web server, you must install pandas, bokeh, flask and flask_bootstrap"
            )
            self._log.debug(
                "Those are not all installed so BAC0 will work in Lite mode only."
            )

        self.response = None
        self._initialized = False
        self._started = False
        self._stopped = False

        if localIPAddr in Base._used_ips:
            raise InitializationError(
                "IP Address provided ({}) already used by BAC0. Check if another software is using port 47808 on this network interface. If so, you can define multiple IP per interface. Or specify another IP using BAC0.lite(ip='IP/mask')".format(
                    localIPAddr
                )
            )

        if validate_ip_address(localIPAddr):
            self.localIPAddr = localIPAddr
        else:
            raise InitializationError(
                "IP Address provided ({}) invalid. Check if another software is using port 47808 on this network interface. If so, you can define multiple IP per interface. Or specify another IP using BAC0.lite(ip='IP/mask')".format(
                    localIPAddr
                )
            )

        self.Boid = (
            int(deviceId) if deviceId else (3056177 + int(random.uniform(0, 1000)))
        )

        self.segmentationSupported = segmentationSupported
        self.maxSegmentsAccepted = maxSegmentsAccepted
        self.localObjName = localObjName
        self.local_objects = LocalObjects(device=self)

        self.maxAPDULengthAccepted = maxAPDULengthAccepted
        self.vendorId = vendorId
        self.vendorName = vendorName
        self.modelName = modelName
        self.description = description

        self.discoveredDevices = None
        self.systemStatus = DeviceStatus(1)

        self.bbmdAddress = bbmdAddress
        self.bbmdTTL = bbmdTTL
        self.bdtable = bdtable

        self.firmwareRevision = firmwareRevision
        self._ric = {}
        self.subscription_contexts = {}

        try:
            self.startApp()
        except InitializationError as error:
            raise InitializationError(
                "Gros probleme : {}. Address requested : {}".format(error, localIPAddr)
            )

    def startApp(self):
        """
        Define the local device, including services supported.
        Once defined, start the BACnet stack in its own thread.
        """
        self._log.debug("Create Local Device")
        try:
            # make a device object
            self.this_device = LocalDeviceObject(
                objectName=self.localObjName,
                objectIdentifier=self.Boid,
                maxApduLengthAccepted=int(self.maxAPDULengthAccepted),
                segmentationSupported=self.segmentationSupported,
                vendorIdentifier=self.vendorId,
                vendorName=self.vendorName,
                modelName=self.modelName,
                systemStatus=self.systemStatus,
                description=self.description,
                firmwareRevision=self.firmwareRevision,
                applicationSoftwareVersion=infos.__version__,
                protocolVersion=1,
                protocolRevision=0,
            )

            # make an application
            if self.bdtable:
                self.this_application = BAC0BBMDDeviceApplication(
                    self.this_device,
                    self.localIPAddr,
                    bdtable=self.bdtable,
                    iam_req=self._iam_request(),
                    subscription_contexts=self.subscription_contexts,
                )
                app_type = "BBMD Device"
            elif self.bbmdAddress and self.bbmdTTL > 0:

                self.this_application = BAC0ForeignDeviceApplication(
                    self.this_device,
                    self.localIPAddr,
                    bbmdAddress=self.bbmdAddress,
                    bbmdTTL=self.bbmdTTL,
                    iam_req=self._iam_request(),
                    subscription_contexts=self.subscription_contexts,
                )
                app_type = "Foreign Device"
            else:
                self.this_application = BAC0Application(
                    self.this_device,
                    self.localIPAddr,
                    iam_req=self._iam_request(),
                    subscription_contexts=self.subscription_contexts,
                )
                app_type = "Simple BACnet/IP App"
            self._log.debug("Starting")
            self._initialized = True
            try:
                self._startAppThread()
                Base._used_ips.add(self.localIPAddr)
                self._log.info("Registered as {}".format(app_type))
            except OSError as error:
                self._log.warning("Error opening socket: {}".format(error))
                raise InitializationError("Error opening socket: {}".format(error))
            self._log.debug("Running")
        except OSError as error:
            self._log.error("an error has occurred: {}".format(error))
            raise InitializationError("Error starting app: {}".format(error))
        finally:
            self._log.debug("finally")

    def register_foreign_device(self, addr=None, ttl=0):
        self.this_application.bip.register(addr, ttl)

    def unregister_foreign_device(self):
        self.this_application.bip.unregister()

    def disconnect(self):
        """
        Stop the BACnet stack.  Free the IP socket.
        """
        self._log.debug("Stopping All running tasks")
        stopAllTasks()
        self._log.debug("Stopping BACnet stack")
        # Freeing socket
        try:
            self.this_application.mux.directPort.handle_close()
        except:
            self.this_application.mux.broadcastPort.handle_close()

        stopBacnetIPApp()  # Stop Core
        self._stopped = True  # Stop stack thread
        self.t.join()
        self._started = False
        Base._used_ips.remove(self.localIPAddr)
        self._log.info("BACnet stopped")

    def _startAppThread(self):
        """
        Starts the BACnet stack in its own thread so requests can be processed.
        As signal cannot be called in another thread than the main thread
        when calling startBacnetIPApp, we must pass None to both parameters
        """
        self._log.info("Starting app...")
        enable_sleeping(0.0005)
        self.t = Thread(
            target=startBacnetIPApp,
            kwargs={"sigterm": None, "sigusr1": None},
            daemon=True,
        )
        try:
            self.t.start()
            self._started = True
            self._log.info("BAC0 started")
        except OSError:
            stopBacnetIPApp()
            self.t.join()
            raise

    @property
    def discoveredNetworks(self):
        return self.this_application.nse._learnedNetworks or set()

    #    @property
    #    def routing_table(self):
    #        return self.this_application.nse._routing_table or {}

    @property
    def routing_table(self):
        """
        Routing Table will give all the details about routers and how they
        connect BACnet networks together.

        It's a decoded presentation of what bacpypes.router_info_cache contains.

        Returns a dict with the address of routers as key.
        """

        class Router:
            def __init__(self, router_info, index=None, path=None):
                self.source_network = router_info.snet
                self.address = router_info.address
                self.destination_networks = router_info.dnets
                self.index = index
                self.path = path

            def __repr__(self):
                return "Source Network: {} | Address: {} | Destination Networks: {} | Path: {}".format(
                    self.source_network,
                    self.address,
                    self.destination_networks,
                    self.path,
                )

        self._routers = {}

        self._ric = {}
        ric = self.this_application.nsap.router_info_cache

        for networks, routers in ric.routers.items():
            for address, router in routers.items():
                self._routers[str(address)] = Router(router, index=networks)
        for path, router in ric.path_info.items():
            self._routers[str(router.address)].path = path

        return self._routers



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
ScheduleWrite.py - creation of ReinitializeDeviceRequest

"""
from ..io.Read import find_reason
from ..io.IOExceptions import (
    SegmentationNotSupported,
    ReadPropertyException,
    ReadPropertyMultipleException,
    NoResponseFromController,
    ApplicationNotStarted,
)
from ...core.utils.notes import note_and_log

# --- standard Python modules ---
import datetime as dt

# --- 3rd party modules ---
from bacpypes.pdu import Address, GlobalBroadcast
from bacpypes.primitivedata import Integer, Date, Time, CharacterString
from bacpypes.basetypes import DateTime
from bacpypes.apdu import WritePropertyRequest, SimpleAckPDU
from bacpypes.iocb import IOCB
from bacpypes.core import deferred
from bacpypes.basetypes import DateTime, DailySchedule, TimeValue, Time
from bacpypes.constructeddata import ArrayOf, Any

from bacpypes.primitivedata import Null, Atomic, Integer, Unsigned, Real, Enumerated

from datetime import time as dt_time
from datetime import datetime as dt


@note_and_log
class Schedule:
    """
    Everything you need to write to a schedule
    """

    WeeklySchedule = ArrayOf(DailySchedule)
    schedules = {}
    days = [
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    ]

    schedule_example_multistate = {
        "states": {"Occupied": 1, "UnOccupied": 2, "Standby": 3, "Not Set": 4},
        "week": {
            "monday": [("1:00", "Occupied"), ("17:00", "UnOccupied")],
            "tuesday": [("2:00", "Occupied"), ("17:00", "UnOccupied")],
            "wednesday": [("3:00", "Occupied"), ("17:00", "UnOccupied")],
            "thursday": [("4:00", "Occupied"), ("17:00", "UnOccupied")],
            "friday": [("5:00", "Occupied"), ("17:00", "UnOccupied")],
            "saturday": [("6:00", "Occupied"), ("17:00", "UnOccupied")],
            "sunday": [("7:00", "Occupied"), ("17:00", "UnOccupied")],
        },
    }
    schedule_example_binary = {
        "states": {"inactive": 0, "active": 1},
        "week": {
            "monday": [("1:00", "active"), ("16:00", "inactive")],
            "tuesday": [("2:00", "active"), ("16:00", "inactive")],
            "wednesday": [("3:00", "active"), ("16:00", "inactive")],
            "thursday": [("4:00", "active"), ("16:00", "inactive")],
            "friday": [("5:00", "active"), ("16:00", "inactive")],
            "saturday": [("6:00", "active"), ("16:00", "inactive")],
            "sunday": [("7:00", "active"), ("16:00", "inactive")],
        },
    }
    schedule_example_analog = {
        "states": "analog",
        "week": {
            "monday": [("1:00", 22), ("18:00", 19)],
            "tuesday": [("2:00", 22), ("18:00", 19)],
            "wednesday": [("3:00", 22), ("18:00", 19)],
            "thursday": [("4:00", 22), ("18:00", 19)],
            "friday": [("5:00", 22), ("18:00", 19)],
            "saturday": [("6:00", 22), ("18:00", 19)],
            "sunday": [("7:00", 22), ("18:00", 19)],
        },
    }

    def create_weeklySchedule(self, dict_schedule, object_reference=None):
        """
        From a structured dict (see schedule_example), create a WeeklySchedule
        an ArrayOf(DailySchedule)
        """
        object_reference = object_reference
        ds = dict_schedule
        if object_reference is not None:
            self.schedules[object_reference] = ds

        def _set_value(v):
            try:
                if dict_schedule["states"].lower() == "analog":
                    return Real(v)
            except AttributeError:
                if dict_schedule["states"] == ["inactive", "active"]:
                    return Integer(dict_schedule["states"][v])
                else:
                    return Integer(dict_schedule["states"][v] - 1)

        daily_schedules = []
        for day in Schedule.days:
            list_of_events = dict_schedule["week"][day]
            _daily_schedule = [
                TimeValue(time=event[0], value=_set_value(event[1]))
                for event in list_of_events
            ]
            daily_schedules.append(DailySchedule(daySchedule=_daily_schedule))
        return Schedule.WeeklySchedule(daily_schedules)

    def make_weeklySchedule_request(self, destination, object_instance, weeklySchedule):
        request = WritePropertyRequest(
            objectIdentifier=("schedule", object_instance),
            propertyIdentifier="weeklySchedule",
        )

        address = Address(destination)
        request.pduDestination = address
        request.propertyValue = Any()
        request.propertyValue.cast_in(weeklySchedule)
        request.priority = 15
        return request

    def send_weeklyschedule_request(self, request):
        iocb = IOCB(request)
        iocb.set_timeout(10)
        deferred(self.this_application.request_io, iocb)

        iocb.wait()

        if iocb.ioResponse:  # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, SimpleAckPDU):  # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug(
                    "Not an ack. | APDU : {} / {}".format((apdu, type(apdu)))
                )
                return
        if iocb.ioError:  # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            raise NoResponseFromController("APDU Abort Reason : {}".format(reason))

        self._log.info(
            "Schedule Write request sent to device : {}".format(request.pduDestination)
        )

    def write_weeklySchedule(self, destination, schedule_instance, schedule):
        weeklyschedule = self.create_weeklySchedule(schedule)
        request = self.make_weeklySchedule_request(
            destination=destination,
            object_instance=schedule_instance,
            weeklySchedule=weeklyschedule,
        )
        self.send_weeklyschedule_request(request)

    def read_weeklySchedule(self, address, schedule_instance):
        """
        This function will turn the weeklySchedule received into a 
        human readable dict. 
        This dict can then be modified and used to write back to controller
        using bacnet.write_weeklySchedule
        """
        try:
            _schedule, object_references, reliability, priority, presentValue = self.readMultiple(
                "{} schedule {} weeklySchedule listOfObjectPropertyReferences reliability priorityForWriting presentValue".format(
                    address, schedule_instance
                )
            )

        except Exception:
            # Rwad Multiple not supported... try single prop read
            def _read(prop):
                return self.read(
                    "{} schedule {} {}".format(address, schedule_instance, prop)
                )

            try:
                _schedule = _read("weeklySchedule")
                object_references = _read("listOfObjectPropertyReferences")
                reliability = _read("reliability")
                priority = _read("priorityForWriting")
                presentValue = _read("presentValue")

            except Exception:
                raise ()

        schedule = {}
        _state_text = None
        offset_MV = 0 if len(object_references) == 0 else 1

        try:
            first_obj_id = object_references[0]
            obj_type, obj_instance = first_obj_id.objectIdentifier
            if obj_type == "binaryValue":
                _state_text = ["inactive", "active"]
            elif "multi" in obj_type:
                _state_text = self.read(
                    "{} {} {} stateText".format(address, obj_type, obj_instance)
                )
            elif "analog" in obj_type:
                _state_text = "analog"

            schedule["object_references"] = [
                objectId.objectIdentifier for objectId in object_references
            ]
            schedule["references_names"] = [
                self.read(
                    "{} {} {} objectName".format(
                        address, each.objectIdentifier[0], each.objectIdentifier[1]
                    )
                )
                for each in object_references
            ]
        except Exception as error:
            self._log.error("Error {}".format(error))
            # Not used in device, not linked...
            _state_text = range(255)
            schedule["object_references"] = []
            schedule["references_names"] = []

        schedule["states"] = {}
        if _state_text == ["inactive", "active"]:
            for i, each in enumerate(_state_text):
                schedule["states"][each] = i
            presentValue = "{} ({})".format(
                list(schedule["states"].keys())[int(presentValue.value)],
                presentValue.value,
            )
        elif _state_text != "analog":
            for i, each in enumerate(_state_text):
                schedule["states"][each] = i + offset_MV
            try:
                presentValue = "{} ({})".format(
                    list(schedule["states"].keys())[
                        int(presentValue.value) - offset_MV
                    ],
                    presentValue.value,
                )
            except TypeError:
                presentValue = presentValue.value
        else:
            schedule["states"] = _state_text
            presentValue = "{}".format(presentValue.value)
        schedule["reliability"] = reliability
        schedule["priority"] = priority
        schedule["presentValue"] = presentValue
        schedule["week"] = self.decode_weeklySchedule(_schedule, _state_text, offset_MV)

        return schedule

    def decode_weeklySchedule(self, weeklySchedule, states, offset_MV):
        week = {}
        for i, day in enumerate(Schedule.days):
            week[day] = self.decode_dailySchedule(weeklySchedule[i], states, offset_MV)
        return week

    def decode_dailySchedule(self, dailySchedule, states, offset_MV):
        events = []
        for each in dailySchedule.daySchedule:
            hour, minute, second, hundreth = each.time
            _time = dt_time(hour, minute, second, hundreth).strftime("%H:%M")
            try:
                if states == ["inactive", "active"]:
                    _value = states[int(each.value.value)]
                elif states == "analog":
                    _value = float(each.value.value)
                else:
                    _value = states[int(each.value.value) - offset_MV]
                events.append((_time, _value))
            except IndexError:
                events.append((_time, "??? ({})".format(each.value.value)))
        return events



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
Points.py - Definition of points so operations on Read results are more convenient.
"""

# --- standard Python modules ---
from datetime import datetime
from collections import namedtuple
import time

# --- 3rd party modules ---
try:
    import pandas as pd
    from pandas.io import sql

    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False

from bacpypes.object import TrendLogObject

# --- this application's modules ---
from ...tasks.Poll import SimplePoll as Poll
from ...tasks.Match import Match, Match_Value
from ..io.IOExceptions import NoResponseFromController, UnknownPropertyError
from ..utils.notes import note_and_log


# ------------------------------------------------------------------------------


class VirtualDeviceProperties(object):
    """
    This serves as a container for device properties
    """

    def __init__(self):
        self.name = "Virtual Device"
        self.address = "local"
        self.device_id = None
        self.network = None
        self.pollDelay = None
        self.objects_list = None
        self.pss = None
        self.multistates = None
        self.db_name = None
        self.segmentation_supported = True
        self.history_size = None
        self.bacnet_properties = None

    def __repr__(self):
        return "{}".format(self.asdict)

    @property
    def asdict(self):
        return self.__dict__


class VirtualDevice(object):
    def __init__(self):
        self.properties = VirtualDeviceProperties()


class VirtualPointProperties(object):
    """
    A container for VirtualPoint properties
    """

    def __init__(self):
        self.name = None
        self.device = VirtualDevice()
        self.description = ""
        self.units_state = ""
        self.type = "virtual"
        self.history_size = None
        self._df = None

    def __repr__(self):
        return "VIRTUAL POINT-- {} | Descr : {}".format(self.name, self.description)


@note_and_log
class VirtualPoint(VirtualPointProperties):
    """
    Virtual points could be used to show calculation result on Bokeh
    A function is passed at creation time and this function must return a pandas Serie
    """

    def __init__(
        self,
        name,
        initial_value=None,
        history_fn=None,
        description=None,
        units="No Units",
    ):
        if description is None:
            raise ValueError("Please provide description")
        if not _PANDAS:
            raise ImportError("Pandas required to use VirtualPoints")
        self.name = name
        self.properties = VirtualPointProperties()
        self.properties.name = name
        self.properties.description = description
        self.properties.units_state = units
        self._history_fn = history_fn

        self._history = namedtuple("_history", ["timestamp", "value"])
        self._history.timestamp = []
        self._history.value = []

        self._match_task = namedtuple("_match_task", ["task", "running"])
        self._match_task.task = None
        self._match_task.running = False

        self.fake_pv = None
        if initial_value:
            self._set(initial_value)

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        self._log.warning("Use bacnet.add_trend(point) instead")

    def _set(self, value):
        if self._history_fn is None:
            self.fake_pv = value
            self._trend(value)
        else:
            self._log.warning(
                "Point is configured as a function of other points. You can't set a new value"
            )

    def _trend(self, res):
        self._history.timestamp.append(datetime.now())
        self._history.value.append(res)
        if self.properties.history_size is None:
            return
        else:
            if self.properties.history_size < 1:
                self.properties.history_size = 1
            if len(self._history.timestamp) >= self.properties.history_size:
                try:
                    self._history.timestamp = self._history.timestamp[
                        -self.properties.history_size :
                    ]
                    self._history.value = self._history.value[
                        -self.properties.history_size :
                    ]
                    assert len(self._history.timestamp) == len(self._history.value)
                except Exception as e:
                    self._log.exception("Can't append to history")

    @property
    def value(self):
        """
        Retrieve value of the point
        """
        return self.lastValue

    @property
    def lastValue(self):
        """
        returns: last value read
        """
        if _PANDAS:
            return self.history.dropna().iloc[-1]
        else:
            return self._history.value[-1]

    @property
    def history(self):
        """
        returns : (pd.Series) containing timestamp and value of all readings
        """
        if self._history_fn is not None:
            return self._history_fn()
        else:
            if not _PANDAS:
                return dict(zip(self._history.timestamp, self._history.value))
            idx = self._history.timestamp.copy()
            his_table = pd.Series(index=idx, data=self._history.value[: len(idx)])
            del idx
            his_table.name = ("{}/{}").format(
                self.properties.device.properties.name, self.properties.name
            )
            his_table.units = self.properties.units_state
            his_table.states = "virtual"

            his_table.description = self.properties.description

            his_table.datatype = self.properties.type
            return his_table

    def match_value(self, value, *, delay=5):
        """
        This allow functions like :
            device['point'].match('value')

        A sensor will follow a calculation...
        """
        if self._match_task.task is None:
            self._match_task.task = Match_Value(value=value, point=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            self._match_task.task.stop()
            self._match_task.running = False
            time.sleep(1)

            self._match_task.task = Match_Value(value=value, point=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError("Stop task before redefining it")

    def __repr__(self):
        return "{}/{} : {:.2f} {}".format(
            self.properties.device.properties.name,
            self.properties.name,
            self.value,
            self.properties.units_state,
        )

    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __eq__(self, other):
        return self.value == other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other
